import requests
import time
import traceback
import datetime
import re
import uuid
import socket


class angelAPI():
    LTPData_per_sec = 1
    MaxQuoteSymbolLimit = 50
    Historical_per_sec = 3

    # BASE_URL = "https://apiconnect.angelbroking.com/" #For V3
    BASE_URL = "https://apiconnect.angelone.in/"
    ENDPOINTS = { # For V3
        "getData" : "/rest/secure/angelbroking/historical/v1/getCandleData",
        "getQuotes" : "rest/secure/angelbroking/market/v1/quote"
    }

    HistoricalMapping = {
        "1" : ["ONE_MINUTE", 30], # angel API has a one minute data ,have 30 days limit
        "3" : ["THREE_MINUTE", 60],
        "5" : ["FIVE_MINUTE", 100],
        "10" : ["TEN_MINUTE", 100],
        "15" : ["FIFTEEN_MINUTE", 200],
        "30" : ["THIRTY_MINUTE", 200],
        "60" : ["ONE_HOUR", 400],
        "1D" : ["ONE_DAY", 2000]
    }


    try:
        clientPublicIp = requests.get('https://api.ipify.org').text.strip()
        hostname = socket.gethostname()
        clientLocalIp = socket.gethostbyname(hostname)
    except Exception as e:
        clientPublicIp = "106.193.147.98"
        clientLocalIp = "127.0.0.1"

    clientMacAddress = ':'.join(re.findall('..', '%012x' % uuid.getnode()))
    accept = "application/json"
    userType = "USER"
    sourceID = "WEB"

    def __init__(self,sessionData : dict):
        self.reqsession = requests.Session()
        self.headers = {
            "Content-type": self.accept,
            "X-ClientLocalIP": self.clientLocalIp,
            "X-ClientPublicIP": self.clientPublicIp,
            "X-MACAddress": self.clientMacAddress,
            "Accept": self.accept,
            "X-PrivateKey": sessionData['ApiKey'],
            "X-UserType": self.userType,
            "X-SourceID": self.sourceID
        }
        self.set_accesstoken(sessionData['Sessionid']['jwtToken'])
        self.__supportedFormats = list(self.info()['supported timeformats']["timeValues"].values())

    def set_accesstoken(self, accesstoken):
        self.headers['Authorization'] = f"Bearer {accesstoken}"

    def requestAPI(self, method, url, body=None, data=None, is_header=True, timeout=15):
        try:
            if is_header:
                resp = self.reqsession.request(method, url, json=body, data=data, timeout=timeout,
                                               headers=self.headers)
            else:
                resp = self.reqsession.request(method, url, json=body, data=data, timeout=timeout)
            data = resp.json()
            if data.get("status"):
                return {"status": True, "data": [data], "error": False, "message": "Data Received"}
            else:
                try:
                    return {"status": False, "data": [], "error": True, "message": data['message']}
                except :
                    return {"status": False, "data": [], "error": True, "message": data}

        except requests.exceptions.Timeout:
            return {"status": False, "data": [], "error": True, "message": "Timeout error"}
        except Exception as e:
            try:
                return {"status": False, "data": [], "error": True, "errorMessage": str(e), "message": resp.text}
            except:
                return {"status": False, "data": [], "error": True, "message": str(e)}


    def get_quotes(self, symbols: list,timesleep=0):
        tokenMap = {}
        exchangeTokens = {}
        nonTokens = []

        # Map tokens to exchange:symbol and categorize tokens
        for token in symbols:
            histToken = token.get('histToken')
            if histToken:
                exchange, symbol = histToken.split(", ")
                tokenMap[f'{exchange}:{symbol}'] = token['token']
                exchangeTokens.setdefault(exchange, []).append(symbol)
            else:
                nonTokens.append(token['token'])

        # Check if there are any valid tokens with histToken
        if not exchangeTokens:
            return {"status": False, "data": [], "error": True, "message": "No histToken found for any symbols"}

        # Make API request
        url = f'{self.BASE_URL}{self.ENDPOINTS["getQuotes"]}'
        params = {"mode": 'FULL', "exchangeTokens": exchangeTokens}
        response = self.requestAPI("POST", url, body=params)

        if not response.get('status') or response.get('error'):
            return response

        # Process API response data
        main_data = {}
        fetched_data = response.get('data', [{}])[0].get('data', {}).get('fetched', [])
        if not fetched_data:
            return {"status": False, "data": [], "error": True, "message": "No data received"}

        for mes in fetched_data:
            symbol = f"{mes['exchange']}:{mes['symbolToken']}"
            feedTime = datetime.datetime.strptime(mes['exchFeedTime'], "%d-%b-%Y %H:%M:%S")
            main_data[tokenMap[symbol]] = {
                "timestamp_str": feedTime.strftime("%Y-%m-%d %H:%M:%S"),
                "timestamp": str(int(feedTime.timestamp())),
                "symbol": tokenMap[symbol],
                "ltp": mes['ltp'],
                "prev_day_close": mes['close'],
                "oi": mes['opnInterest'],
                "prev_day_oi": 0,
                "turnover": mes['tradeVolume'],
                "best_bid_price": mes['depth']['buy'][0]['price'],
                "best_ask_price": mes['depth']['sell'][0]['price'],
                "best_bid_qty": mes['depth']['buy'][0]['quantity'],
                "best_ask_qty": mes['depth']['sell'][0]['quantity'],
                "ttq": mes['tradeVolume'],
                "token": tokenMap[symbol]
            }

        # Include empty data for tokens without histToken
        for token in nonTokens:
            main_data[token] = {}

        return {"status": True, "data": main_data, "error": False, "message": "Data Received"}

    def getHistorical(self, token, startTime: datetime, endTime: datetime, interval: str):
        if interval not in self.HistoricalMapping:
            return {"status": False, "error": True, "message": f"Invalid interval. Please use: {self.info()}", "data": []}

        if not token:
            return {"status": False, "data": [], "error": True, "message": "No histToken found for the symbol"}

        exchange, symboltoken = token.split(", ")
        main_data = []
        endTime = min(endTime, datetime.datetime.now())

        while startTime < endTime:
            ed = min(startTime + datetime.timedelta(days=self.HistoricalMapping[interval][1] - 1), endTime)

            params = {
                "exchange": exchange,
                "symboltoken": symboltoken,
                "interval": interval[0],
                "fromdate": startTime.strftime("%Y-%m-%d %H:%M"),
                "todate": ed.strftime("%Y-%m-%d %H:%M")
            }

            response = self.requestAPI("POST", f"{self.BASE_URL}{self.ENDPOINTS['getData']}", body=params)

            if not response.get('status') or response.get('error'):
                return response

            candle_data = response.get('data', [])
            if candle_data:
                main_data.extend(self.ConvertToSTD(candle_data[0].get('data', [])))
            else:
                return {"status": False, "data": [], "error": True, "message": "Data not available"}

            startTime = ed

        return {"status": True, "data": main_data, "error": False, "message": "Data Received"} if main_data else \
            {"status": False, "data": [], "error": True, "message": "Data not available"}

    def ConvertToSTD(self,data):
        if not data:
            return []
        std_format = [
            {
                'datetime': datetime.datetime.strptime(candle[0], "%Y-%m-%dT%H:%M:%S%z").strftime("%Y-%m-%d %H:%M:%S"),
                'open': float(candle[1]),
                'high': float(candle[2]),
                'low': float(candle[3]),
                'close': float(candle[4]),
                'volume': int(candle[5])
            }
            for candle in data
        ]
        return std_format

    def info(self):
        return {"supported timeformats" :
            {"timeValues" :
                {
                    "1 minute" : "1",
                    "3 minute" : "3",
                    "5 minute"  : "5",
                    "10 minute" : "10",
                    "15 minute" : "15",
                    "30 minute" : "30",
                    "60 minute" : "60",
                    "1 Day" : "1D"
                }}}



from .iiflxtsws import IIFLXTSWS
from .fyersws import FyersWs
from .angelws import AngelWs
from .kotakws import Kotakws
from .shoonyaws import ShoonyaWS
from .upstoxWS import UpstoxWS
from ...sparkLib import SparkLib



from maticalgos.zmqWsHandler import wsLib, dataStream
from maticalgos.sparkLib import SparkLib

ac = SparkLib(apikeys="Api Keys")
ac.generate_token()

tokens = ac.getTokens("NIFTY", "NFO")
tokens = [f"{i['exch_seg']}:{i['token']}" for i in tokens['data']]
tokens = tokens[:300]

ws = wsLib()

ws.connect("Account Name")
ws.subscribe("Account Name", tokens)

ws.reconnect("Account Name")
ws.stop("Account Name")

############# Stream Data 

def ltp(message):
    print(message)
    
dt = dataStream()
dt.tickStream=ltp 

dt.connect()
#Subscribe to the tokens required in the strategy
# dt.subscribe(tokens)

# dt.stop() 




import datetime
import traceback
import json 
import os
from .dependencies import KotakDataFeed, SparkLib

class Kotakws():
    dataStore = {}
    BROKER = "KOTAKNEO"
    TOKENS = {}
    def __init__(self, accountData, accessToken = None, dataType = "ltp", 
                 onLtp = None, 
                 onDepth = None, 
                 onError = None, 
                 onClose = None,
                 onOpen = None,
                 ) :
        self.accountData = accountData
        self.dType = False if dataType == "ltp" else True if dataType == "DepthUpdate" else None #Currently Only supports LTP Updates.
        self.onLtp = self.__onLTP if not onLtp  else onLtp
        self.onDepth = self.__onDepth if not onDepth else onDepth
        self.onError = self.__onError if not onError else onError
        self.onClose = self.__onClose if not onClose else onClose
        self.onOpen = self.__onOpen if not onOpen else onOpen
        self.run = True
        if not accessToken :
            if os.environ.get("MATICALGOS_AccessToken"):
                accessToken = os.environ["MATICALGOS_AccessToken"] 
            else: 
                raise Exception("Please generate access token.")
        self.spk = SparkLib(access_token=accessToken)
    
    def __onLTP(self, message):
        print(message)
    
    def __onDepth(self, message):
        print(message)
    
    def __onError(self, message):
        print("error : ",message)

    def __onClose(self, *args):
        print("Connection Closed")

    def __onOpen(self):
        print("Connection Opened")
    
    def getTokens(self, tokens):
        data = self.spk.getBrokerTokens(tokens=tokens, broker=self.BROKER)
        if data['status'] and not data['error'] : 
            tokens = data['data']
            return tokens
        else: 
            raise Exception (data['message'])
    
    def getSymbol(self, exchange, token):
        return self.TOKENS[", ".join([exchange, token])]

    def tickHandler(self, message):
        try:
            if message.get('data') != None : 
                for dt in message['data'] : 
                    if dt['name'] == "sf" : 
                        data = {
                            "time":dt.get('fdtm') if dt.get('fdtm') != None else dt.get("ltt"),
                            "bidPrice" : dt.get('bp') if dt.get('bp') != None else dt.get('ltp') if dt.get('ltp') != None else 0 ,
                            "askPrice" : dt.get('sp') if dt.get('sp') != None else dt.get('ltp') if dt.get('ltp') != None else 0,
                            "bidQty" : dt.get('bq') if dt.get('bq') != None else 0,
                            "askQty" : dt.get('bs') if dt.get('bs') != None else 0,
                            "token" : dt.get('tk'),
                            "exchange" : dt.get('e'),
                            "ltp" : dt.get('ltp') if dt.get('ltp') != None else (float(dt['bp'])+float(dt['sp']))/2 if None not in [dt.get('bp'),dt.get('sp')] else None
                                }
                        if None not in data.values() : 
                            dtime = datetime.datetime.strptime(data['time'], "%d/%m/%Y %H:%M:%S")
                            token = self.getSymbol(data['exchange'], data['token'])
                            openInterest = 0 if token.split(":")[0] not in ['NFO',"BFO","MCX"]  else dt.get('oi')
                            if openInterest == None :
                                try:
                                    openInterest = self.dataStore[token]['oi'] if self.dataStore.get(token) != None else 0
                                except:
                                    openInterest = 0
                            turnover = dt.get('to') if dt.get('to') != None else \
                                        self.dataStore[token]['to'] if self.dataStore.get(token) != None and self.dataStore[token].get('to') != None \
                                        else 0
                            turnover = 0 if turnover == None else turnover   
                            # volume = dt.get('v') if dt.get('v') != None else data[token]['v'] if data.get(token) != None else 0 
                            volume = dt.get('v') if dt.get('v') != None and dt.get('v') != 0 else \
                                        self.dataStore[token]['v'] if self.dataStore.get(token) != None and self.dataStore[token].get('v') != None \
                                        else 0 
                            volume = 0 if volume == None else volume
                            prevClose = dt.get('c') if dt.get('c') != None \
                                        else float(data.get('ltp'))-float(dt.get('cng')) if dt.get('cng') != None \
                                        else self.dataStore.get(token)['c'] if self.dataStore.get(token) != None and self.dataStore[token].get('c') != None else \
                                        0.0
                                        
                            # if self.dataStore.get(token) != None :
                            #     try: 
                            #         prevClose = self.dataStore[token]['c']
                            #     except:
                            #         prevClose = 0.0
                            # else: 
                            #     prevClose = dt.get('c') if dt.get('c') != None else float(data['ltp'])-float(dt.get('cng')) if dt.get('cng') != None else 0.0
                            msg = {
                                "timestamp_str" : str(dtime),
                                "timestamp" : str(int(dtime.timestamp())),
                                "symbol" : token, 
                                "ltp" : float(data['ltp']), 
                                "prev_day_close" : prevClose, 
                                "oi" : int(openInterest), 
                                "prev_day_oi" : 0, 
                                "turnover" : float(turnover) , 
                                "best_bid_price" : float(data['bidPrice']), 
                                "best_ask_price" : float(data['askPrice']), 
                                "best_bid_qty" : int(data["bidQty"]),
                                "best_ask_qty" : int(data['askQty']),
                                "ttq" : int(volume),  
                                "token" : token
                                }
                            self.onLtp(msg)
                            self.dataStore[token] = {"c" : prevClose, 'oi' : openInterest, 'to' : turnover, "v" : volume}
                        else: 
                            if None not in  [dt.get('tk'),dt.get('e')]:
                                token = self.getSymbol(dt.get('e'), dt.get('tk'))
                                prevclose = dt.get('c') if dt.get('c') != None else float(dt.get('ltp'))-float(dt.get('cng')) if (dt.get('cng') != None and dt.get("ltp") != None) else None
                                data = {
                                        "c" : prevclose,
                                        "oi" : dt.get('oi'),
                                        "v" : dt.get("v") ,
                                        'to' : dt.get('to')
                                    }
                                if self.dataStore.get(token) != None : 
                                    data = {k : data[k] for k in data if data[k] != None }
                                    self.dataStore[token].update(data)
                                else: 
                                    data = {k : 0 for k in data if data[k] == None }
                                    self.dataStore[token] = data

                    elif dt['name'] == 'if' : 
                        data = {
                            "time":dt.get('tvalue'),
                            "volume": 0,
                            "ltp" : dt.get('iv'),
                            "bidPrice" : 0,
                            "askPrice" : 0,
                            "bidQty" : 0,
                            "askQty" : 0,
                            "turnOver" : 0,
                            "token" : dt.get('tk'),
                            "exchange" : dt.get('e'),
                                }
                        if None not in data.values() : 
                            dtime = datetime.datetime.strptime(data['time'], "%d/%m/%Y %H:%M:%S")
                            token = self.getSymbol(data['exchange'], data['token'])
                            prevClose = dt.get('ic') if dt.get('ic') != None else float(data["ltp"])-float(dt.get('cng')) if dt.get('cng') == None else None
                            if prevClose == None and self.dataStore.get(token) != None :
                                prevClose = self.dataStore[token]['c']
                            else: 
                                prevClose = 0.0

                            msg = {
                                    "timestamp_str" : str(dtime),
                                    "timestamp" : str(int(dtime.timestamp())),
                                    "symbol" : token, 
                                    "ltp" : float(data['ltp']), 
                                    "prev_day_close" : prevClose, 
                                    "oi" : 0, 
                                    "prev_day_oi" : 0, 
                                    "turnover" : data['turnOver'] , 
                                    "best_bid_price" : data['bidPrice'], 
                                    "best_ask_price" : data['askPrice'], 
                                    "best_bid_qty" : data["bidQty"],
                                    "best_ask_qty" : data['askQty'],
                                    "ttq" : data['volume'],  
                                    "token" : token
                                    }
                                
                            self.dataStore[token] = {"c" : prevClose, 'oi' : 0}
                            self.onLtp(msg)
                        
        except Exception as e : 
            self.onError(f"Error : {e}, data : {message}, Traceback : {traceback.format_exc()}")

    def disconnect(self):
        self.run = False
        self.soc.close_websocket()
    
    def chunk_list(self, data, chunk_size):
        return [data[i:i + chunk_size] for i in range(0, len(data), chunk_size)]

    def Subscribe(self, tokens):
        try: 
            tokens = self.getTokens(tokens)
            toSubs = []
            toSubsIndex = []
            notFound = []
            for tk in tokens:
                wsToken = tk['wsToken'] 
                if wsToken != None : 
                    exch, wstk = wsToken.split(", ")
                    self.TOKENS[wsToken] = tk['token']
                    if tk['is_index'] :
                        toSubsIndex.append({"instrument_token" : wstk, "exchange_segment" : exch})
                    else: 
                        toSubs.append({"instrument_token" : wstk, "exchange_segment" : exch})
                else: 
                    notFound.append(tk['token'])

            if toSubs != [] : 
                chunked_list = self.chunk_list(toSubs,99)
                for lst in chunked_list:
                    self.soc.get_live_feed(lst, isIndex=False, isDepth=False)
            if toSubsIndex != [] : 
                chunked_list = self.chunk_list(toSubsIndex,99)
                for lst in chunked_list:
                    self.soc.get_live_feed(lst, isIndex=True, isDepth=False) 
            if notFound != [] : 
                self.onError(f"Error : Unable to find tokens for {notFound} ")

        except Exception as e :
            self.onError(f"Error : {str(e)}, Traceback : {traceback.format_exc()}")

    def Unsubscribe(self, tokens):
        try: 
            tokens = self.getTokens(tokens)
            toSubs = []
            toSubsIndex = []
            notFound = []
            for tk in tokens:
                wsToken = tk['wsToken'] 
                if wsToken != None : 
                    exch, wstk = wsToken.split(", ")
                    if tk['is_index'] :
                        toSubsIndex.append({"instrument_token" : wstk, "exchange_segment" : exch})
                    else: 
                        toSubs.append({"instrument_token" : wstk, "exchange_segment" : exch})
                else: 
                    notFound.append(tk['token'])

            if toSubs != [] : 
                chunked_list = self.chunk_list(toSubs,99)
                for lst in chunked_list:
                    self.soc.un_subscribe_list(lst, isIndex=False, isDepth=False)
            if toSubsIndex != [] : 
                chunked_list = self.chunk_list(toSubsIndex,99)
                for lst in chunked_list:
                    self.soc.un_subscribe_list(lst, isIndex=True, isDepth=False) 
            if notFound != [] : 
                self.onError(f"Error : Unable to find tokens for {notFound} ")

        except Exception as e :
            self.onError(f"Error : {str(e)}, Traceback : {traceback.format_exc()}")

    def connect(self):
        try: 
            sessionData = self.accountData['Sessionid']
            sessionData = sessionData if type(sessionData) == dict else json.loads(sessionData)
            self.soc = KotakDataFeed(token=sessionData['token'],
                                     sid=sessionData['sid'],
                                     server_id=sessionData['hsServerId'],
                                     on_message=self.tickHandler,
                                     on_close=self.onClose,
                                     on_error=self.onError,
                                     on_open=self.onOpen,
                                     retry_attempts=5)
            self.soc.start_websocket()

        except Exception as e : 
            self.onClose(f"Error : {e}, Traceback : {traceback.format_exc()}")



import json
import threading
import websocket
import logging
from time import sleep

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def reportmsg(msg):
    # print(msg)
    logger.debug(msg)

class FeedType:
    TOUCHLINE = 1
    SNAPQUOTE = 2

class NorenApi:
    __service_config = {
        'websocket_endpoint': 'wss://api.shoonya.com/NorenWSTP/',
    }

    def __init__(self,
                 message_update_callback = None,
                 socket_open_callback = None,
                 socket_close_callback = None,
                 socket_error_callback = None,
                 susertoken = None,
                 userid = None,
                 retry_attempts = 5,
                 ):
        self.__websocket = None
        self.__websocket_connected = False
        self.__ws_mutex = threading.Lock()
        self.__on_error = None
        self.__on_disconnect = None
        self.__on_open = None
        self.__message_callback = None
        self.__order_update_callback = None
        self.__subscribers = {}
        self.__market_status_messages = []
        self.__exchange_messages = []
        self.max_retry_attempts = 5
        self.current_retry_attempt = 0
        self.retry_delay_seconds = 5
        self.__username   = userid
        self.__accountid  = userid
        self.__susertoken = susertoken
        self.max_retry_attempts = retry_attempts if retry_attempts > -1 else self.max_retry_attempts
        self.__on_open = socket_open_callback
        self.__on_disconnect = socket_close_callback
        self.__on_error = socket_error_callback
        self.__message_callback = message_update_callback

    def __ws_send(self, *args, **kwargs):
        while self.__websocket_connected == False:
            sleep(0.05)  # sleep for 50ms if websocket is not connected, wait for reconnection
        with self.__ws_mutex:
            ret = self.__websocket.send(*args, **kwargs)
        return ret

    def __on_close_callback(self, wsapp, close_status_code, close_msg):
        reportmsg(close_status_code)
        reportmsg(wsapp)

        self.__websocket_connected = False
        if self.__on_disconnect:
            self.__on_disconnect()

    def __on_open_callback(self, ws=None):
        self.__websocket_connected = True
        #prepare the data
        values              = { "t": "c" }
        values["uid"]       = self.__username
        values["actid"]     = self.__username
        values["susertoken"]    = self.__susertoken
        values["source"]    = 'API'

        payload = json.dumps(values)
        reportmsg(payload)
        self.__ws_send(payload)
        self.current_retry_attempt = 0
        #self.__resubscribe()


    def __on_error_callback(self, ws=None, error=None):
        if(type(ws) is not websocket.WebSocketApp): # This workaround is to solve the websocket_client's compatiblity issue of older versions. ie.0.40.0 which is used in upstox. Now this will work in both 0.40.0 & newer version of websocket_client
            error = ws
        if self.__on_error:
            self.__on_error(error)
        self.retry_connect()

    def __on_data_callback(self, ws=None, message=None, data_type=None, continue_flag=None):
        res = json.loads(message)
        # print(res)
        if(self.__message_callback is not None):
            if res['t'] == 'tk' or res['t'] == 'tf':
                self.__message_callback(res)
                return
            if res['t'] == 'dk' or res['t'] == 'df':
                self.__message_callback(res)
                return

        if(self.__on_error is not None):
            if res['t'] == 'ck' and res['s'] != 'OK':
                self.__on_error(res)
                return

        if self.__on_open:
            if res['t'] == 'ck' and res['s'] == 'OK':
                self.__on_open()
                return

    def ws_run_forever(self):
        self.__websocket.run_forever( ping_interval=3,  ping_payload='{"t":"h"}')#, ping_timeout=9)

    def start_websocket(self):
        """ Start a websocket connection for getting live data """
        url = self.__service_config['websocket_endpoint'].format(access_token=self.__susertoken)
        reportmsg('connecting to {}'.format(url))
        try:
            self.__websocket = websocket.WebSocketApp(url,
                                                      on_data=self.__on_data_callback,
                                                      on_error=self.__on_error_callback,
                                                      on_close=self.__on_close_callback,
                                                      on_open=self.__on_open_callback,
                                                      )
            # self.__websocket.run_forever( ping_interval=2,  ping_payload='{"t":"h"}', ping_timeout=1)
            self.__ws_thread = threading.Thread(target=self.ws_run_forever)
            self.__ws_thread.daemon = False
            self.__ws_thread.start()

        except Exception as e:
            logger.error(f"Error starting websocket: {e}")
            self.retry_connect()

    def close_websocket(self):
        if self.__websocket_connected == False:
            return
        self.__websocket_connected = False
        self.__websocket.close()
        self.__ws_thread.join()
        self.__ws_thread = None

    def subscribe(self, instrument, feed_type=FeedType.TOUCHLINE):
        values = {}

        if(feed_type == FeedType.TOUCHLINE):
            values['t'] =  't'
        elif(feed_type == FeedType.SNAPQUOTE):
            values['t'] =  'd'
        else:
            values['t'] =  str(feed_type)

        if type(instrument) == list:
            values['k'] = '#'.join(instrument)
        else :
            values['k'] = instrument

        data = json.dumps(values)

        #print(data)
        self.__ws_send(data)

    def unsubscribe(self, instrument, feed_type=FeedType.TOUCHLINE):
        values = {}

        if(feed_type == FeedType.TOUCHLINE):
            values['t'] =  'u'
        elif(feed_type == FeedType.SNAPQUOTE):
            values['t'] =  'ud'

        if type(instrument) == list:
            values['k'] = '#'.join(instrument)
        else :
            values['k'] = instrument

        data = json.dumps(values)

        #print(data)
        self.__ws_send(data)

    def retry_connect(self):
        if self.current_retry_attempt < self.max_retry_attempts:
            self.current_retry_attempt += 1
            sleep(self.retry_delay_seconds)
            print(f"Retrying connection Attempt {self.current_retry_attempt} of {self.max_retry_attempts}")
            self.start_websocket()
        else:
            logger.error("Max retry attempts reached. Exiting")
            self.close_websocket()



from maticalgos.sparkLib import OrderSocket
# from sparkLib import SparkLib
import time

def on_order(message):
    print('test message', message)

def on_error(error):
    print('test_error', error)

def on_close():
    print("### closed ###")

def on_open():
    print(" test on_open called")

order = OrderSocket(access_token='<Your Access Token>',
                    on_order=on_order,
                    on_error=on_error,
                    on_close=on_close,
                    on_connect=on_open,
                    reconnect=True,
                    max_reconnect_attempts=20,
                    run_background=True
                    )
order.connect()
time.sleep(20)
order.close_connection()

###############
from maticalgos.sparkLib.utility import IIFLXTSWS
from maticalgos.sparkLib import SparkLib

spk = SparkLib(apikeys="")
spk.generate_token()

accountData = spk.getOneAccount(accountName= "AccountName")

if accountData["status"] and not accountData['error']:
    accountData = accountData["data"][0]
else: 
    raise Exception(accountData)

def onLTP(message):
    print(message)

def onError(message):
    print(message)

def onClose(message):
    print(message)

xts = IIFLXTSWS(accountData=accountData,
                onLtp=onLTP,
                onError=onError,
                onClose=onClose)

xts.connect()

# Test as per tokens
xts.Subscribe(["NSE:212", "NSE:26009", "NFO:1234", "BSE:2134", "BFO:5678"])

# Close connection 
xts.closeConnection()




import requests
import time
import traceback
import datetime


class FyersHist():
    LTPData_per_sec = 10
    MaxQuoteSymbolLimit = 50

    BASE_URL = "https://api-t1.fyers.in/"
    ENDPOINTS = { # For V3
        "getData" : "data/history?symbol={fyers_token}&resolution={resolution}&date_format={date_format}&range_from={start_date}&range_to={end_date}&cont_flag={cont_flag}&oi_flag={oi_flag}",
        "getQuotes" : "data/quotes?symbols={SYMBOLS}"
    }

    HistoricalMapping = {
        "5S" : ['5S', 30],  # Fyers has 5 seconds data of 30 days limit
        "10S" : ['10S', 30],
        "15S" : ['15S', 30],
        "30S" : ['30S', 30],
        "45S" : ['45S', 30],
        "1" : ['1', 100],
        "2" : ['2', 100],
        "3" : ['3', 100],
        "5" : ['5', 100],
        "10" : ['10', 100],
        "15" : ['15', 100],
        "20" : ['20', 100],
        "30" : ['30', 100],
        "60" : ['60', 100],
        "120" : ['120', 100],
        "240" : ['240', 100],
        "1D" : ['1D', 365]
    }


    def __init__(self,sessionData : dict):
        self.appid = sessionData['ApiKey']
        self.authcode = sessionData['Sessionid']
        self.reqsession = requests.Session()
        self.header = {
                "Accept" : "*/*",
                "Accept-Encoding" : "gzip, deflate, br",
                "Connection" : "keep-alive", 
                "Version" : "2.1"
            }
        self.set_accesstoken()
        self.__supportedFormats = list(self.info()['supported timeformats']["timeValues"].values())

    def set_accesstoken(self):
        self.header['Authorization'] = f"{self.appid}:{self.authcode}"

    def requestAPI(self, method, url, body = None, is_header = True, timeout = 15):
        try :
            if not is_header :
                resp = self.reqsession.request(method, url, json = body, timeout = timeout)
            else :
                resp = self.reqsession.request(method, url, json = body, timeout = timeout,
                                               headers = self.header)

            data = resp.json()

            if data.get('s') == 'error' :
                return {"status" : False, "data" : [data], "error" : True}

            else :
                return {"status" : True, "data" : [data], "error" : False}

        except requests.exceptions.Timeout:
            return {"status" : False, "data" : [], "error" : True, "message" : "Timeout error"}

        except Exception as e :
            return {"status" : False, "data" : [], "error" : True, "message" : e}

    def get_quotes(self, symbols: list,sleeptime = 0):
        main_data = {}
        tokenMap = {}
        nonTokens = []

        # Map tokens to exchange:symbol and categorize tokens
        for token in symbols:
            histToken = token.get('histToken')
            if histToken:
                tokenMap[histToken] = token['token']
            else:
                nonTokens.append(token['token'])

        if not tokenMap:
            return {
                "status": False,
                "data": [],
                "error": True,
                "message": "No histToken found for any symbols"
            }

        tokenMapKeys = list(tokenMap.keys())

        for i in range(0, len(tokenMapKeys), self.MaxQuoteSymbolLimit):
            j = min(i + self.MaxQuoteSymbolLimit, len(tokenMapKeys))
            symbols_str = ",".join(tokenMapKeys[i:j])
            url = f'{self.BASE_URL}{self.ENDPOINTS["getQuotes"]}'.format(SYMBOLS=symbols_str)

            try:
                response = self.requestAPI("GET", url)
                if not response.get('status') or response.get('error'):
                    return response

                api_data = response.get('data', [])
                if not api_data or not api_data[0].get('d'):
                    return {
                        "status": False,
                        "data": [],
                        "error": True,
                        "message": "No data received"
                    }

                for message in api_data[0]['d']:
                    symbol = message['n']
                    message_values = message['v']
                    main_data[tokenMap[symbol]] = {
                        "timestamp_str": datetime.datetime.fromtimestamp(
                            int(message_values['tt'])
                        ).replace(hour=0, minute=0, second=0, microsecond=0).strftime('%Y-%m-%d %H:%M:%S'),
                        "timestamp": str(message_values['tt']),
                        "symbol": tokenMap[symbol],
                        "ltp": message_values['lp'],
                        "prev_day_close": message_values['prev_close_price'],
                        "oi": 0,
                        "prev_day_oi": 0,
                        "turnover": message_values['volume'],
                        "best_bid_price": message_values['bid'],
                        "best_ask_price": message_values['ask'],
                        "best_bid_qty": 0,
                        "best_ask_qty": 0,
                        "ttq": message_values['volume'],
                        "token": tokenMap[symbol]
                    }

            except Exception as e:
                return {
                    "status": False,
                    "data": [],
                    "error": True,
                    "message": f"Error occurred: {str(e)}"
                }
            time.sleep(sleeptime)

        # Fill in empty data for nonTokens
        for token in nonTokens:
            main_data[token] = {}

        return {
            "status": True,
            "data": main_data,
            "error": False,
            "message": "Data Received"
        }

    def getHistorical(self, token, startTime : datetime, endTime : datetime, interval : str):
        if interval not in self.__supportedFormats:
            return {"status" : False, "error" : True, "message" : f"Invalid interval. Please input as per the format: {self.info()}", "data" : []}

        if not token:
            return {"status" : False, "data" : [], "error" : True, "message" : "No histToken found for the symbol"}

        endTime = min(endTime, datetime.datetime.now())
        main_data = []

        while startTime < endTime:
            ed = min(startTime + datetime.timedelta(days=self.HistoricalMapping[interval][1]-1), endTime)

            url = f'{self.BASE_URL}{self.ENDPOINTS["getData"]}'.format(fyers_token=token, resolution=self.HistoricalMapping[interval][0], date_format=0, start_date=int(startTime.timestamp()), end_date=int(ed.timestamp()), cont_flag=1, oi_flag=1)
            response = self.requestAPI("GET", url)
            if not response.get('status') or response.get('error'):
                return response

            candle_data = response.get('data', [])

            if candle_data:
                main_data.extend(self.ConvertToSTD(candle_data[0].get('candles', [])))
            else:
                return {"status": False, "data": [], "error": True, "message": "Data not available"}

            startTime = ed

        return {"status": True, "data": main_data, "error": False, "message": "Data Received"} if main_data else \
                {"status": False, "data": [], "error": True, "message": "Data not available"}

    def ConvertToSTD(self, data):
        if not data:
            return []

        std_format = [
            {
                'datetime': datetime.datetime.fromtimestamp(candle[0]).strftime("%Y-%m-%d %H:%M:%S"),
                'open': float(candle[1]),
                'high': float(candle[2]),
                'low': float(candle[3]),
                'close': float(candle[4]),
                'volume': int(candle[5]),
                'oi': int(candle[6]) if len(candle) == 7 else 0
            }
            for candle in data
        ]

        return std_format

    def info(self):
        return {"supported timeformats" :
            {"timeValues" :
                {
                    "5 seconds" : "5S",
                    "10 seconds" : "10S",
                    "15 seconds" : "15S",
                    "30 seconds" : "30S",
                    "45 seconds" : "45S",
                    "1 minute" : "1",
                    "2 minute" : "2",
                    "3 minute" : "3",
                    "5 minute"  : "5",
                    "10 minute" : "10",
                    "15 minute" : "15",
                    "20 minute" : "20",
                    "30 minute" : "30",
                    "60 minute" : "60",
                    "120 minute" : "120",
                    "240 minute" : "240",
                    "1 Day" : "1D"
                }}}




from .XTSdataSocket import XTSMDSocket_io
from .XTSdataApi import XTS_MarketData
from .fyersAPIv3.data_ws import FyersDataSocket
from .fyersAPIv3.data_ws import defines
from .fyersAPIv3.data_ws import FyersLogger
from .smartDataWS import SmartWebSocketV2
from .kotakData import KotakDataFeed
from .shoonyaDataWS import NorenApi as ShoonyaDataWS
from .UpstoxDataws import AsyncUpstockOrderUpdate
from ....sparkLib import SparkLib


import datetime
import traceback
import asyncio
import os 
import queue
from .dependencies import AsyncUpstockOrderUpdate
from .dependencies import SparkLib

class UpstoxWS():
    BROKER = "UPSTOX"
    TOKENS = {}
    
    def __init__(self, accountData, accessToken = None, dataType = "ltp", 
                 onLtp = None, 
                 onDepth = None, 
                 onError = None, 
                 onClose = None,
                 onOpen = None,
                 ) : 
        self.accountData = accountData
        self.dType = "full"  # if dataType == "ltp" else "full" if dataType == "DepthUpdate" else None #Currently Only supports LTP Updates.
        self.onLtp = self.__onLTP if not onLtp  else onLtp
        self.onDepth = self.__onDepth if not onDepth else onDepth
        self.onError = self.__onError if not onError else onError
        self.onClose = self.__onClose if not onClose else onClose
        self.onOpen = self.__onOpen if not onOpen else onOpen
        self.run = True
        if not accessToken :
            if os.environ.get("MATICALGOS_AccessToken"):
                accessToken = os.environ["MATICALGOS_AccessToken"] 
            else: 
                raise Exception("Please generate access token.")
        self.spk = SparkLib(access_token=accessToken)

    def __onLTP(self, message):
        print(message)
    
    def __onDepth(self, message):
        print(message)
    
    def __onError(self, message):
        print("error : ",message)

    def __onClose(self, *args):
        print("Connection Closed")

    def __onOpen(self):
        print("Connection Opened")

    async def on_message(self, message):
        self.tickHandler(message)

    async def on_connect(self):
        self.onOpen()
        # await client.subscribe("ltpc", ['NSE_FO|36805'])

    async def on_error(self, error):
        print("WebSocket error:", error)
        self.onError(error)

    async def on_close(self, code, reason):
        self.onClose(code, reason)

    def getTokens(self, tokens):
        data = self.spk.getBrokerTokens(tokens=tokens, broker=self.BROKER)
        if data['status'] and not data['error'] : 
            tokens = data['data']
            return tokens
        else: 
            raise Exception (data['message'])
    
    def getSymbol(self, token):
        return self.TOKENS[token]
    
    def tickHandler(self, message):
        for key in message['feeds'].keys():
            try: 
                token = self.getSymbol(key)
                data = message['feeds'][key]['ff']
                dKey = next(iter(data))
                if dKey == "indexFF" : 
                    tickData = data[dKey]
                    msg = {
                            "timestamp_str" : str(datetime.datetime.fromtimestamp(int(tickData['ltpc']['ltt'])/1000)),
                            "timestamp" :tickData['ltpc']['ltt'],
                            "symbol" : token, 
                            "ltp" : float(tickData['ltpc']['ltp']), 
                            "prev_day_close" : float(tickData['ltpc']['cp']), 
                            "oi" : 0, 
                            "prev_day_oi" : 0, 
                            "turnover" : 0 , 
                            "best_bid_price" : 0, 
                            "best_ask_price" : 0, 
                            "best_bid_qty" : 0,
                            "best_ask_qty" : 0,
                            "ttq" : 0,  
                            "token" : token
                                        }
                    self.onLtp(msg)
                elif dKey == "marketFF":
                    tickData = data[dKey]
                    msg = {
                            "timestamp_str" : str(datetime.datetime.fromtimestamp(int(tickData['ltpc']['ltt'])/1000)),
                            "timestamp" :tickData['ltpc']['ltt'],
                            "symbol" : token, 
                            "ltp" : float(tickData['ltpc']['ltp']), 
                            "prev_day_close" : float(tickData['ltpc']['cp']), 
                            "oi" : int(tickData['eFeedDetails']['oi']) if tickData['eFeedDetails'].get('oi') else 0, 
                            "prev_day_oi" : int(tickData['eFeedDetails']['poi']) if tickData['eFeedDetails'].get('poi') else 0, 
                            "turnover" : int(tickData['eFeedDetails']['vtt']) if tickData['eFeedDetails'].get('vtt') else 0, 
                            "best_bid_price" : float(tickData['marketLevel']['bidAskQuote'][0]['bp']), 
                            "best_ask_price" : float(tickData['marketLevel']['bidAskQuote'][0]['ap']), 
                            "best_bid_qty" : float(tickData['marketLevel']['bidAskQuote'][0]['bq']),
                            "best_ask_qty" : float(tickData['marketLevel']['bidAskQuote'][0]['aq']),
                            "ttq" : int(tickData['eFeedDetails']['vtt']) if tickData['eFeedDetails'].get('vtt') else 0,  
                            "token" : token
                                        }
                    self.onLtp(msg)
            except Exception as e :
                self.onError(f"Error : {str(e)}, Traceback : {traceback.format_exc()}")

    def Subscribe(self, tokens):
        try: 
            tokens = self.getTokens(tokens)
            toSubs = []
            notFound = []
            for tk in tokens:
                wsToken = tk['wsToken'] 
                if wsToken != None : 
                    self.TOKENS[wsToken] = tk['token']
                    toSubs.append(wsToken)
                else: 
                    notFound.append(tk['token'])
            if toSubs != [] : 
                self.subsQueue.put_nowait({'func':"sub","token":toSubs})
                # self.client.subscribe(self.dType, toSubs)
            if notFound != [] : 
                self.onError(f"Error : Unable to find tokens for {notFound} ")

        except Exception as e :
            self.onError(f"Error : {str(e)}, Traceback : {traceback.format_exc()}")

    def Unsubscribe(self, tokens):
        try: 
            tokens = self.getTokens(tokens)
            toSubs = []
            notFound = []
            for tk in tokens:
                wsToken = tk['wsToken'] 
                if wsToken != None : 
                    toSubs.append(wsToken)
                else: 
                    notFound.append(tk['token'])

            if toSubs != [] : 
                self.subsQueue.put_nowait({'func':"unsub","token":toSubs})
                # self.client.unsubscribe(self.dType, toSubs)
            if notFound != [] : 
                self.onError(f"Error : Unable to find tokens for {notFound} ")

        except Exception as e :
            self.onError(f"Error : {str(e)}, Traceback : {traceback.format_exc()}")

    def disconnect(self):
        self.run = False

    async def _connect(self):
        try : 
            self.client = AsyncUpstockOrderUpdate(
                                                auth_token=self.accountData['Sessionid'],
                                                on_message=self.on_message,
                                                on_connect=self.on_connect,
                                                on_error=self.on_error,
                                                on_close=self.on_close
                                            )
            asyncio.create_task(self.client.connect())
            while self.run : 
                if not self.subsQueue.empty():
                    data = await self.subsQueue.get()
                    if data['func'] == "sub": 
                        await self.client.subscribe(self.dType, data['token'])
                    elif data['func'] == "unsub":
                        await self.client.unsubscribe(self.dType, data['token'])

                await asyncio.sleep(1)
            await self.client.close_connection()
        
        except Exception as e : 
            self.onClose(f"Error : {e}, Traceback : {traceback.format_exc()}")

    def connect(self): 
        self.subsQueue = asyncio.Queue()
        asyncio.run(self._connect())



# -*- coding: utf-8 -*-
"""
Created on Tue Dec 19 12:44:34 2023

@author: Niraj
"""

import requests

ENDPOINTS = {
    'Login':'/auth/login',
    'Quote':'/instruments/quotes',
    'Subscription':'/instruments/subscription', # Same for Unsubscription just PUT instead of POST 
    'indexList' : "/instruments/indexlist?exchangeSegment={exch}" ,
    "optSymbol" :  "/instruments/instrument/optionSymbol?exchangeSegment={exch}&series={series}&symbol={symbol}&expiryDate={exp}&optionType={optype}&strikePrice={strike}",
    "eqSymbol" : "/instruments/instrument/symbol?exchangeSegment={exch}&series={series}&symbol={symbol}",
    "futSymbol" : "/instruments/instrument/futureSymbol?exchangeSegment={exch}&series={series}&symbol={symbol}&expiryDate={exp}"
}

class XTS_MarketData:
    
    def __init__(self, apikey = None, apisecret = None, sessionid = None, URL = None):
        self.BASE_URL = URL
        self.LoginSource = "WebAPI"
        self.apikey = apikey
        self.apisecret = apisecret
        self.reqsession = requests.Session()
        self.header = {'Content-Type': 'application/json'}
        if sessionid != None: 
            self.set_accesstoken(sessionid)
        
    def login(self):
        url = self.BASE_URL + ENDPOINTS["Login"]
        params = {
            "appKey": self.apikey,
            "secretKey": self.apisecret,
            "source": self.LoginSource,
            }
        resp = self.requestAPI("POST", url, body= params)
        if resp['status'] : 
            resp['message'] = "Data Received"
            self.header['Authorization'] = resp['data'][0]['result']['token']
        else : 
            resp['message'] = "Data Received"
        return resp
            
    def set_accesstoken(self, sessionid):
        self.header['Authorization'] = sessionid
        
    def quote(self, instrument):
        params = {'instruments': instrument, 'xtsMessageCode': 1501, 'publishFormat': "JSON"}
        url = self.BASE_URL + ENDPOINTS["Quote"]
        return self.requestAPI("POST", url, body = params)
    
    def subscribe(self, Instruments, xtsMessageCode):
        params = {'instruments': Instruments, 'xtsMessageCode': xtsMessageCode}
        url = self.BASE_URL + ENDPOINTS["Subscription"]
        return self.requestAPI("POST", url, body = params)
    
    def unsubscribe(self, Instruments, xtsMessageCode):
        params = {'instruments': Instruments, 'xtsMessageCode': xtsMessageCode}
        url = self.BASE_URL + ENDPOINTS["Subscription"]
        return self.requestAPI("PUT", url, body = params)

    def getIndexList(self, exchange):
        url = self.BASE_URL + ENDPOINTS["indexList"].format(exch = exchange)
        return self.requestAPI("GET", url)
    
    def optSymbol(self, exchange, series, symbol, exp, opType, strikePrice):
        url = self.BASE_URL + ENDPOINTS['optSymbol'].format(exch = exchange, series = series, symbol = symbol, exp = exp, optype = opType, strike = strikePrice)
        return self.requestAPI("GET", url)
    
    def equitySymbol(self, exchange, series, symbol):
        url = self.BASE_URL + ENDPOINTS['eqSymbol'].format(exch = exchange, series = series, symbol = symbol)
        return self.requestAPI("GET", url)
    
    def futSymbol(self, exchange, series, symbol, exp):
        url = self.BASE_URL + ENDPOINTS['futSymbol'].format(exch = exchange, series = series, symbol = symbol, exp = exp)
        return self.requestAPI("GET", url)
    
    def requestAPI(self, method, url, body = None, is_header = True, timeout = 15): 
        try : 
            if not is_header : 
                resp = self.reqsession.request(method, url, json = body, timeout = timeout)
            else :
                resp = self.reqsession.request(method, url, json = body, timeout = timeout, 
                                               headers = self.header)
            data = resp.json()
            if data.get('type') == 'error' : 
                return {"status" : False, "data" : [data], "error" : True, "statuscode" : resp.status_code}
            
            else : 
                return {"status" : True, "data" : [data], "error" : False, "statuscode" : resp.status_code}
        
        except requests.exceptions.Timeout:
            return {"status" : False, "data" : [], "error" : True,
                    "statuscode" : "resp.status_code", "message" : "Timeout error"}
        
        except Exception as e : 
            return {"status" : False, "data" : [], "error" : True, 
                    "statuscode" : "resp.status_code", "message" : e}


import base64
import logging
import threading
import time
from typing import Optional, Callable

from pkg_resources import resource_filename
import requests
import urllib.parse
import websocket
from threading import Thread
import struct
import json
import os 

# from fyers_apiv3.FyersWebsocket import defines
# from fyers_apiv3.fyers_logger import FyersLogger
from . import defines
from .fyers_logger import FyersLogger 

class SymbolConversion:
    def __init__(self, access_token: str, data_type: str, log_path: str):
        """
        Initializes a SymbolConversion instance.

        Args:
            access_token (str): The access token used for authentication.
            data_type (str): The data_type associated with the symbol conversion. 
                            Valid values are 'Symbolupdate' or 'DepthUpdate'.

        """
        self.data_type = data_type
        if ":" in access_token:
            access_token = access_token.split(":")[1]
        self.access_token = access_token
        self.log_path = log_path
        self.symbols_token_api = "https://api-t1.fyers.in/data/symbol-token"

        if log_path:
            self.log_path = log_path + "/"
        else:
            self.log_path = ""
        self.data_logger = FyersLogger(
            "FyersDataSocket",
            "DEBUG",
            stack_level=3,
            logger_handler=logging.FileHandler(log_path + "fyersDataSocket.log"),
        )

    def symbol_to_hsmtoken(self, symbols: list):
        """
        Converts symbols to HSM tokens.

        Args:
            symbols (list): A list of symbols to be converted.

        Returns:
            tuple: A tuple containing dictionary and list.
                - The first dictionary represents the mapping of symbols to HSM tokens.
                - The second list represents any symbols that could not be converted.

        """
        try:
            data = {"symbols": symbols}
            response = requests.post(
                url=self.symbols_token_api ,
                headers={
                    "Authorization": self.access_token,
                    "Content-Type": "application/json",
                },json=data
            )
            response_data =response.json()
            datadict = {}
            project_path =  os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))
            file_path = os.path.join(project_path, 'dependencies', 'fyersAPIv3', 'map.json')
            # file_path = resource_filename('fyers_apiv3.FyersWebsocket', 'map.json')
            with open(file_path, "r") as file:
                mapper = json.load(file)
            index_dict = mapper["index_dict"]
            exch_seg_dict = mapper["exch_seg_dict"]
            wrong_symbol = []
            dp_index_flag = False

            if response_data['s'] == "ok":
                for symbol, fytoken in response_data["validSymbol"].items():
                    ex_sg = fytoken[:4]
                    if ex_sg not in exch_seg_dict:
                        continue
                    segment = exch_seg_dict[ex_sg]
                    symbol_split = symbol.split("-")
                    update_dict = True
                    if len(symbol_split) > 1 and symbol_split[-1] == "INDEX" and self.data_type != "DepthUpdate":
                        if symbol in index_dict:
                                exch_token = index_dict[symbol]
                        else:
                            exch_token = (
                                symbol.split(":")[1].split("-")[0]
                            )
                        hsm_symbol = (
                                "if" + "|" + segment + "|" + exch_token
                            )                
                    elif self.data_type == "DepthUpdate" and symbol_split[-1] != "INDEX":
                        
                        exch_token = fytoken[10:]
                        hsm_symbol = (
                            "dp" + "|" + segment + "|" + exch_token
                        )
                    elif self.data_type == "SymbolUpdate":
                        exch_token = fytoken[10:]
                        hsm_symbol = (
                            "sf" + "|" + segment + "|" + exch_token
                        )                        
                    elif self.data_type == "DepthUpdate" and symbol_split[-1] == "INDEX":
                        update_dict = False
                        dp_index_flag =True

                    if update_dict:
                        datadict[hsm_symbol] = symbol
                if response_data["invalidSymbol"]:
                    wrong_symbol = response_data["invalidSymbol"]
                return (datadict, wrong_symbol, dp_index_flag,"")

            elif response_data['s'] == "error":
                
                return ({}, [],dp_index_flag, response_data["message"])


        except Exception as e:
            self.data_logger.exception(e)


class FyersDataSocket:

    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(
        self,
        access_token: str,
        write_to_file: bool = False,
        log_path: Optional[str] = None,
        litemode: bool = False,
        reconnect: bool = True,
        on_message: Optional[Callable] = None,
        on_error: Optional[Callable] = None,
        on_connect: Optional[Callable] = None,
        on_close: Optional[Callable] = None,
        reconnect_retry: int = 5
    ):

        """
        Initialize YourClass instance.

        Args:
            access_token (str): The access token used for authentication.
            write_to_file (bool, optional): Specifies if the class should run in the background.
                Defaults to False.
            log_path (str, optional): The path to the log file. Defaults to None.
            litemode (bool, optional): Specifies if the class should run in litemode.
                Defaults to False.
            reconnect (bool, optional): Specifies if the class should attempt to reconnect on failure.
                Defaults to True.
            on_message (callable, optional): Callback function to be executed when a message is received.
                Defaults to None.
            on_error (callable, optional): Callback function to be executed when an error occurs.
                Defaults to None.
            on_connect (callable, optional): Callback function to be executed when a connection is established.
                Defaults to None.
            on_close (callable, optional): Callback function to be executed when a connection is closed.
                Defaults to None.
        """
        
        self.__url = "wss://socket.fyers.in/hsm/v1-5/prod"
        self.__access_token = access_token
        self.__hsm_token = ""
        self.log_path = log_path
        self.lite = litemode
        self.max_retry = reconnect_retry
        self.source = "PythonSDK-3.0.9"
        self.channel_num = 11
        self.channels = []
        self.running_channels = set()
        self.data_type = None
        self.OnMessage = on_message
        self.OnError = on_error
        self.OnOpen = on_connect
        self.OnClose = on_close
        self.UpdateTick = False
        self.ack_count = 0
        self.__ws_run = None
        self.write_to_file = write_to_file
        self.background_flag = False
        self.update_count = 0
        self.literesp = {}
        self.channel_symbol = []
        self.symbol_dict = {}
        self.scrips_count = {}
        self.scrips_per_channel = {}
        self.restart_flag = reconnect
        self.websocket_lock = threading.Lock()
        self.message_lock = threading.Lock()
        self.message_condition = threading.Condition(lock=self.message_lock)
        self.unsub_symbol = []
        for i in range(1, 31):
            self.scrips_per_channel[i] = []
        self.active_channel = None
        self.message = []
        self.resp = {}
        self.__ws_object = None
        self.__valid_token = False
        self.dp_sym = {}
        self.ping_thread = None
        self.message_thread = None
        self.message_thread_stop_event = None
        self.ws_thread = None
        self.infy_loop = None
        self.symbol_limit = 5000
        self.reconnect_attempts = 0
        self.max_reconnect_attempts = 50
        if reconnect_retry < self.max_reconnect_attempts:
            self.max_reconnect_attempts = reconnect_retry
       
        self.mode = "P"
        self.reconnect_delay = 0
        self.index_sym = {}
        self.scrips_sym = {}
        self.symbol_token = {}
        self.ack_bool = False     
        if log_path:
            self.log_path = log_path + "/"
        else:
            self.log_path = ""
        self.data_logger = FyersLogger(
            "FyersDataSocket",
            "DEBUG",
            stack_level=3,
            logger_handler=logging.FileHandler(self.log_path + "fyersDataSocket.log"),
        )
        project_path =  os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))
        file_path = os.path.join(project_path, 'dependencies', 'fyersAPIv3', 'map.json')

        with open(file_path, "r") as file:
            # Imported json file
            mapper = json.load(file)

        self.data_val = mapper["data_val"]
        self.index_val = mapper["index_val"]
        self.lite_val = mapper["lite_val"]
        self.depthvalue = mapper["depthvalue"]

    def access_token_to_hsmtoken(self) -> bool :
        """
        Decode APIv2 token to extract 'hsm_key' and check for validity.

        This function decodes the APIv2 access token and extracts the 'hsm_key' from it.
        It also verifies if the token is expired by comparing the 'exp' (expiration) claim in the
        token's payload with the current timestamp. If the token is valid and not expired, it sets
        the 'hsm_token' attribute and returns True. Otherwise, it raises an error and returns False.

        Returns:
            bool: True if the token is valid and not expired, False otherwise.
        """
        try:
            header_token , payload_b64, _ = self.__access_token.split(".")
            # Decode the base64 encoded payload
            decoded_header = base64.urlsafe_b64decode(header_token + "===")
            decoded_payload = base64.urlsafe_b64decode(payload_b64 + "===")
            # Convert the decoded payload to a string (assuming it's in JSON format)
            decode_token = json.loads(decoded_payload.decode())
            today = int(time.time())
            if decode_token["exp"] - today  < 0:
                self.On_error(
                    {   
                        "type": defines.AUTH_TYPE,
                        "code": defines.TOKEN_EXPIRED,
                        "message": defines.TOKEN_EXPIRED_MSG,
                        "s": defines.ERROR,
                    }
                )
                return False
            self.__valid_token = True

            self.__hsm_token = decode_token["hsm_key"]
            return True
        
        except:
            self.On_error(
                {   "type": defines.AUTH_TYPE,
                    "code": defines.INVALID_CODE,
                    "message": defines.INVALID_TOKEN,
                    "s": defines.ERROR,
                }
            )
            # self.data_logger.error(e)
            return False



    def __access_token_msg(self) -> bytearray:
        """
        Create a message in bytearray for token.

        Returns:
            bytearray: The token message in bytearray format.
        """
        try:
            buffer_size = 18 + len(self.__hsm_token) + len(self.source)

            # Create the byte buffer
            byte_buffer = bytearray()

            # Pack data length into the byte buffer
            byte_buffer.extend(struct.pack("!H", buffer_size - 2))

            # Set ReqType
            byte_buffer.extend(bytes([1]))

            # Set FieldCount
            byte_buffer.extend(bytes([4]))

            # Field-1: AuthToken
            field1_id = 1
            field1_size = len(self.__hsm_token)
            byte_buffer.extend(bytes([field1_id]))
            byte_buffer.extend(struct.pack("!H", field1_size))
            byte_buffer.extend(self.__hsm_token.encode())

            # Field-2
            field2_id = 2
            field2_size = 1
            byte_buffer.extend(bytes([field2_id]))
            byte_buffer.extend(struct.pack("!H", field2_size))
            byte_buffer.extend(self.mode.encode('utf-8'))

            # Field-3
            field3_id = 3
            field3_size = 1
            byte_buffer.extend(bytes([field3_id]))
            byte_buffer.extend(struct.pack("!H", field3_size))
            byte_buffer.extend(bytes([1]))

            # Field-4: self.source
            field4_id = 4
            field4_size = len(self.source)
            byte_buffer.extend(bytes([field4_id]))
            byte_buffer.extend(struct.pack("!H", field4_size))
            byte_buffer.extend(self.source.encode())

            return byte_buffer

        except Exception as e:
            self.data_logger.exception(e)

    def __lite_mode_msg(self) -> bytearray:
        """
        Create a message in bytearray for lite mode connection.

        Returns:
            bytearray: The lite mode message in bytearray format.
        """

        try:
            self.channels = [self.channel_num]
            data = bytearray()

            data.extend(struct.pack(">H", 0))

            data.extend(struct.pack("B", 12))

            data.extend(struct.pack("B", 2))

            channel_bits = 0
            for channel_num in self.channels:
                if channel_num < 64 and channel_num > 0:
                    channel_bits |= 1 << channel_num
            # Field-1
            field_1 = bytearray()
            field_1.extend(struct.pack("B", 1))
            field_1.extend(struct.pack(">H", 8))
            field_1.extend(struct.pack(">Q", channel_bits))
            data.extend(field_1)

            # Field-2
            field_2 = bytearray()
            field_2.extend(struct.pack("B", 2))
            field_2.extend(struct.pack(">H", 1))
            field_2.extend(struct.pack("B", 76))
            data.extend(field_2)

            return data

        except Exception as e:
            self.data_logger.exception(e)

    def __full_mode_msg(self) -> bytearray:

        """
        Create a message in bytearray for full mode connection.

        Returns:
            bytearray: The full mode message in bytearray format.
        """
        try:
            self.channels = [self.channel_num]
            data = bytearray()

            data.extend(struct.pack(">H", 0))

            data.extend(struct.pack("B", 12))

            data.extend(struct.pack("B", 2))

            channel_bits = 0
            for channel_num in self.channels:
                if channel_num < 64 and channel_num > 0:
                    channel_bits |= 1 << channel_num
            # Field-1
            field_1 = bytearray()
            field_1.extend(struct.pack("B", 1))
            field_1.extend(struct.pack(">H", 8))
            field_1.extend(struct.pack(">Q", channel_bits))
            data.extend(field_1)

            # Field-2
            field_2 = bytearray()
            field_2.extend(struct.pack("B", 2))
            field_2.extend(struct.pack(">H", 1))
            field_2.extend(struct.pack("B", 70))
            data.extend(field_2)

            return data

        except Exception as e:
            self.data_logger.exception(e)

    def __subscription_msg(self, symbols: list) -> bytearray:

        """
        Create a message in bytearray for symbol subscription.

        Args:
            symbols (list): A list of symbols to subscribe to.

        Returns:
            bytearray: The subscription message in bytearray format.
        """

        try:
            self.scrips_per_channel[self.channel_num] += symbols
            self.scrips = symbols
            self.scrips_data = bytearray()
            self.scrips_data.append(len(self.scrips) >> 8 & 0xFF)
            self.scrips_data.append(len(self.scrips) & 0xFF)
            for scrip in self.scrips:
                scrip_bytes = str(scrip).encode("ascii")
                self.scrips_data.append(len(scrip_bytes))
                self.scrips_data.extend(scrip_bytes)

            data_len = (
                18 + len(self.scrips_data) + len(self.__access_token) + len(self.source)
            )
            request_type = 4
            field_count = 2
            buffer_msg = bytearray()
            buffer_msg.extend(struct.pack(">H", data_len))
            buffer_msg.append(request_type)
            buffer_msg.append(field_count)

            # Field-1
            buffer_msg.append(1)
            buffer_msg.extend(struct.pack(">H", len(self.scrips_data)))
            buffer_msg.extend(self.scrips_data)

            # Field-2
            buffer_msg.append(2)
            buffer_msg.extend(struct.pack(">H", 1))
            buffer_msg.append(self.channel_num)
            return buffer_msg

        except Exception as e:
            self.data_logger.exception(e)

    def __unsubscription_msg(self, symbols: list) -> bytearray:
        """
        Create a message in bytearray for unsubscription message.

        Args:
            symbols (list): A list of symbols to unsubscribe from.

        Returns:
            bytearray: The unsubscription message in bytearray format.
        """
        try:
            scrips_data = bytearray()
            scrips_data.append(len(symbols) >> 8 & 0xFF)
            scrips_data.append(len(symbols) & 0xFF)
            for scrip in symbols:
                scrip_bytes = str(scrip).encode("ascii")
                scrips_data.append(len(scrip_bytes))
                scrips_data.extend(scrip_bytes)

            data_len = (
                18 + len(scrips_data) + len(self.__access_token) + len(self.source)
            )
            request_type = 5
            field_count = 2
            buffer_msg = bytearray()
            buffer_msg.extend(struct.pack(">H", data_len))
            buffer_msg.append(request_type)
            buffer_msg.append(field_count)

            # Field-1
            buffer_msg.append(1)
            buffer_msg.extend(struct.pack(">H", len(scrips_data)))
            buffer_msg.extend(scrips_data)

            # Field-2
            buffer_msg.append(2)
            buffer_msg.extend(struct.pack(">H", 1))
            buffer_msg.append(self.channel_num)

            return buffer_msg
        except Exception as e:
            self.data_logger.exception(e)

    def __channel_resume_msg(self, channel: int) -> bytearray:

        """
        Create a message in bytearray for channel resume.

        Args:
            channel (int): The channel to resume.

        Returns:
            bytearray: The channel resume message in bytearray format.
        """
        try:

            self.channels = [channel]

            data = bytearray()

            data.extend(struct.pack(">H", 0))

            data.extend(struct.pack("B", 8))

            data.extend(struct.pack("B", 1))

            channel_bits = 0
            for channel_num in self.channels:
                if channel_num < 64 and channel_num > 0:
                    channel_bits |= 1 << channel_num
            # Field-1
            field_1 = bytearray()
            field_1.extend(struct.pack("B", 1))
            field_1.extend(struct.pack(">H", 8))
            field_1.extend(struct.pack(">Q", channel_bits))
            data.extend(field_1)

            return data

        except Exception as e:
            self.data_logger.exception(e)

    def __channel_pause_msg(self, channel: int) -> bytearray:

        """
        Create a message in bytearray for channel pause.

        Args:
            channel (int): The channel to pause.

        Returns:
            bytearray: The channel pause message in bytearray format.
        """

        try:
            self.channels = [channel]

            data = bytearray()

            data.extend(struct.pack(">H", 0))

            data.extend(struct.pack("B", 7))

            data.extend(struct.pack("B", 1))

            channel_bits = 0
            for channel_num in self.channels:
                if channel_num < 64 and channel_num > 0:
                    channel_bits |= 1 << channel_num
            # Field-1
            field_1 = bytearray()
            field_1.extend(struct.pack("B", 1))
            field_1.extend(struct.pack(">H", 8))
            field_1.extend(struct.pack(">Q", channel_bits))
            data.extend(field_1)

            return data

        except Exception as e:
            self.data_logger.exception(e)

    def __ackowledgement_msg(self, message_number: int) -> bytearray:

        """
        Create a message in bytearray for acknowledgement.

        Args:
            message_number (int): The message number to acknowledge.

        Returns:
            bytearray: The acknowledgement message in bytearray format.
        """
        try:
            total_size = 11
            req_type = 3
            field_count = 1
            field_id = 1
            field_size = 4
            field_value = message_number
            buffer_msg = bytearray()
            # Pack the data into the byte array
            buffer_msg.extend(struct.pack(">H", total_size - 2))
            buffer_msg.extend(struct.pack("B", req_type))
            buffer_msg.extend(struct.pack("B", field_count))
            buffer_msg.extend(struct.pack("B", field_id))
            buffer_msg.extend(struct.pack(">H", field_size))
            buffer_msg.extend(struct.pack(">I", field_value))

            return buffer_msg

        except Exception as e:
            self.data_logger.exception(e)

    def __auth_resp(self, data: bytearray) -> dict:
        """
        Unpacks the authentication response from a bytearray.

        Args:
            data (bytearray): The authentication response message.

        Returns:
            dict: The authentication response as a dictionary with keys 'code', 'message', and 's'.
        """

        try:
            offset = 4
            offset += 1
            field_length = struct.unpack("!H", data[offset : offset + 2])[0]
            offset += 2
            string_val = data[offset : offset + field_length].decode("utf-8")
            offset += field_length

            if string_val == "K":

                self.On_message(
                    {   "type": defines.AUTH_TYPE,
                        "code": defines.SUCCESS_CODE,
                        "message": defines.AUTH_SUCCESS,
                        "s": defines.SUCCESS,
                    }
                )
            else:
                self.On_error(
                    {   "type": defines.AUTH_TYPE,
                        "code": defines.AUTH_ERROR_CODE,
                        "message": defines.AUTH_FAIL,
                        "s": defines.ERROR,
                    }
                )

            offset += 1
            field_length = struct.unpack("!H", data[offset : offset + 2])[0]
            offset += 2
            self.ack_count = struct.unpack(">I", data[offset : offset + 4])[0]
            offset += 4

        except Exception as e:
            self.data_logger.exception(e)

    def __subscribe_resp(self, data: bytearray) -> dict:
        """
        Unpacks the subscription response from a bytearray.

        Args:
            data (bytearray): The subscription response message.

        Returns:
            dict: The subscription response as a dictionary with keys 'code', 'message', and 's'.
        """

        try:

            offset = 5
            field_length = struct.unpack("H", data[offset : offset + 2])[0]
            offset += 2
            string_val = data[offset : offset + 1].decode("latin-1")
            offset += field_length
            if string_val == "K":

                self.On_message(
                    {   
                        "type": defines.SUBS_TYPE,
                        "code": defines.SUCCESS_CODE,
                        "message": defines.SUBSCRIBE_SUCCESS,
                        "s": defines.SUCCESS,
                    }
                )
            else:
                self.On_error(
                    {
                        "type": defines.SUBS_TYPE,
                        "code": defines.SUBS_ERROR_CODE,
                        "message": defines.SUBSCRIBE_FAIL,
                        "s": defines.ERROR,
                    }
                )

        except Exception as e:
            self.data_logger.exception(e)

    def __unsubscribe_resp(self, data: bytearray) -> dict:

        """
        Unpacks the unsubscription response from a bytearray.

        Args:
            data (bytearray): The unsubscription response message.

        Returns:
            dict: The unsubscription response as a dictionary with keys 'code', 'message', and 's'.
        """

        try:

            offset = 5
            field_length = struct.unpack("H", data[offset : offset + 2])[0]
            offset += 2
            string_val = data[offset : offset + 1].decode("latin-1")
            offset += field_length
            if string_val == "K":

                self.On_message(
                    {
                        "type": defines.UNSUBS_TYPE,
                        "code": defines.SUCCESS_CODE,
                        "message": defines.UNSUBSCRIBE_SUCCESS,
                        "s": defines.SUCCESS,
                    }
                )
                for symbol in self.unsub_symbol:
                    count = 0
                    for channel in self.running_channels:
                        if symbol in self.scrips_per_channel[channel]:
                            count +=1 
                        if count > 1:
                            break
                    if symbol in self.scrips_per_channel[self.active_channel]:
                        self.scrips_per_channel[self.active_channel].remove(symbol)
                    if count == 1:
                        self.symbol_token.pop(symbol)
                            
            else:
                self.On_error(
                    {
                        "type": defines.UNSUBS_TYPE,
                        "code": defines.UNSUBS_ERROR_CODE,
                        "message": defines.UNSUBSCRIBE_FAIL,
                        "s": defines.ERROR,
                    }
                )

        except Exception as e:
            self.data_logger.exception(e)

    def __lite_full_mode_resp(self, data: bytearray) -> dict:

        """
        Unpacks the lite/full mode response from a bytearray.

        Args:
            data (bytearray): The lite/full mode response message.

        Returns:
            dict: The lite/full mode response as a dictionary with keys 'code', 'message', and 's'.
        """

        try:
            offset = 3

            # Unpack the field count
            field_count = struct.unpack("!B", data[offset : offset + 1])[0]
            offset += 1

            if field_count >= 1:
                # Unpack the field ID
                offset += 1

                # Unpack the field length
                field_length = struct.unpack("!H", data[offset : offset + 2])[0]
                offset += 2

                # Extract the string value and decode it
                string_val = data[offset : offset + field_length].decode("utf-8")
                offset += field_length

                if string_val == "K":
                    if self.lite:
                        self.On_message(
                            {
                                "type": defines.LITE_MODE_TYPE,
                                "code": defines.SUCCESS_CODE,
                                "message": defines.LITE_MODE,
                                "s": defines.SUCCESS,
                            }
                        )
                    else:
                        self.On_message(
                            {
                                "type": defines.FULL_MODE_TYPE,
                                "code": defines.SUCCESS_CODE,
                                "message": defines.FULL_MODE,
                                "s": defines.SUCCESS,
                            }
                        )
                else:
                    self.On_error(
                        {
                            "code": defines.MODE_ERROR_CODE,
                            "message": defines.MODE_CHANGE_ERROR,
                            "s": defines.ERROR,
                        }
                    )

        except Exception as e:
            self.data_logger.exception(e)
        
    def __resume_pause_resp(self, data: bytearray, channeltype: int) -> dict:
        """
        Unpacks and processes the resume/pause response data based on the channel type.

        Args:
            data (bytearray): The response data.
            channeltype (int): The channel type. 7 for pause and 8 for resume.

        Returns:
            dict: The resume/pause response as a dictionary with keys 'code', 'message', and 's'.
        """
        try:
            offset = 5

            # Unpack the field length
            field_length = struct.unpack("!H", data[offset : offset + 2])[0]
            offset += 2

            # Extract the string value and decode it
            string_val = data[offset : offset + field_length].decode("utf-8")
            offset += field_length

            if string_val == "K":
                if channeltype == 7:
                    self.On_message(
                        {
                            "type": defines.CH_PAUSE_TYPE,
                            "code": defines.SUCCESS_CODE,
                            "message": defines.CHANNEL_PAUSED,
                            "s": defines.SUCCESS,
                        }
                    )
                else:
                    self.On_message(
                        {
                            "type": defines.CH_RESUME_TYPE,
                            "code": defines.SUCCESS_CODE,
                            "message": defines.CHANNEL_RESUMED,
                            "s": defines.SUCCESS,
                        }
                    )
            else:
                if channeltype == 7:
                    self.On_error(
                        {
                            "type": defines.CH_PAUSE_TYPE,
                            "code": defines.PAUSE_ERROR_CODE,
                            "message": defines.CHANNEL_CHANGE_FAIL,
                            "s": defines.ERROR,
                        }
                    )
                else:
                    self.On_error(
                        {
                            "type": defines.CH_RESUME_TYPE,
                            "code": defines.RESUME_ERROR_CODE,
                            "message": defines.CHANNEL_CHANGE_FAIL,
                            "s": defines.ERROR,
                        }
                    )

        except Exception as e:
            self.data_logger.exception(e)

    def __response_output(self, data: str, data_type: str) -> object:

        """
        Processes the response data and returns the output based on the specified data_type.

        Args:
            data (bytearray): The response data.
            data_type (str): The type of data to be processed.

        Returns:
            object: The processed output based on the specified data_type.
        """
        try:
            data_resp = data
            precision_calcu_value = [
                                    "ltp",
                                    "bid_price",
                                    "ask_price",
                                    "avg_trade_price",
                                    "low_price",
                                    "high_price",
                                    "open_price",
                                    "prev_close_price",
                                ]
            response = {}
            if (
                "bidPrice1" not in data_resp
                # and "vol_traded_today" in data_resp
                and self.lite
            ):
                for i, val in enumerate(self.lite_val):
                    if val in data_resp and val == "ltp":
                        response[val] = data_resp[val] / (
                                (10 ** data_resp["precision"] ) * data_resp["multiplier"]
                            )
                    else:
                        response[val] = data_resp[val]
                if "prev_close_price" in response and "ltp" in response:
                    response["ch"] = round((response['ltp']  - response['prev_close_price']),2) 
                    response["chp"] = round((response["ch"]  / response['prev_close_price'] * 100) , 2)
            else:
                if data_type == "depth":

                    for i, val in enumerate(self.depthvalue):
                        if val in data_resp and i < 10:
                            response[val] = data_resp[val] / ((
                                10 ** data_resp["precision"] ) * data_resp["multiplier"])

                        elif val in data_resp:
                            response[val] = data_resp[val]
                elif data_type == "scrips":
                    for i, val in enumerate(self.data_val):
                        if val in data_resp and val in precision_calcu_value and val not in ["upper_ckt", "lower_ckt"]:
                            response[val] = data_resp[val] / (
                                (10 ** data_resp["precision"] )* data_resp["multiplier"]
                            )
                            # response[val] = data_resp[val] / (

                        elif val in data_resp:
                            response[val] = data_resp[val]

                    response["lower_ckt"] = 0
                    response["upper_ckt"] = 0
                    if "prev_close_price" in response and "ltp" in response and response["prev_close_price"] != 0:
                        response["ch"] = round((response['ltp']  - response['prev_close_price']),4) 
                        response["chp"] = round((response["ch"]  / response['prev_close_price'] * 100) , 4)
                    if "OI" in response:
                        response.pop("OI")
                    if "Yhigh" in response:
                        response.pop("Yhigh")
                    if "Ylow" in response:
                        response.pop("Ylow")
                else:
                    for i, val in enumerate(self.index_val):
                        if val in data_resp and i in [0, 1, 3, 4, 5]:
                            response[val] = data_resp[val] / (
                                (10 ** data_resp["precision"] ) * data_resp["multiplier"]
                            )
                        elif val in data_resp:
                            response[val] = data_resp[val]
                        if "prev_close_price" in response and "ltp" in response:
                            response["ch"] = round((response['ltp']  - response['prev_close_price']),2) 
                            response["chp"] = round((response["ch"]  / response['prev_close_price'] * 100) , 2)
            
            self.On_message(response)

        except Exception as e:
            self.data_logger.exception(e)

    def __datafeed_resp(self, data: bytearray):

        """
        Unpacks and processes the data based on data_type and sends it to the __response_output function.

        Args:
            data (bytearray): The response data.

        Returns:
            None
        """
        try:

            if self.ack_count > 0:
                self.update_count += 1
                message_num = struct.unpack(">I", data[3:7])[0]
                if self.update_count == self.ack_count:
                    self.ack_msg = self.__ackowledgement_msg(message_num)
                    # self.message.append(self.ack_msg)
                    self.add_message(self.ack_msg)

                    
                    self.update_count = 0
            scrip_count = struct.unpack("!H", data[7:9])[0]
            offset = 9

            for _ in range(scrip_count):
                data_type = struct.unpack("B", data[offset : offset + 1])[0]
                if data_type == 83:  # Snapshot datafeed

                    offset += 1
                    topic_id = struct.unpack("H", data[offset : offset + 2])[0]
                    offset += 2
                    topic_name_len = struct.unpack("B", data[offset : offset + 1])[0]
                    offset += 1

                    topic_name = data[offset : offset + topic_name_len].decode("utf-8")
                    offset += topic_name_len

                    # Maintaining dict - topic_id : topic_name
                    if topic_name[:2] == "dp":
                        self.dp_sym[topic_id] = topic_name

                        self.resp[self.dp_sym[topic_id]] = {}

                        field_count = struct.unpack("B", data[offset : offset + 1])[0]
                        offset += 1

                        for index in range(field_count):
                            value = struct.unpack(">i", data[offset : offset + 4])[0]
                            offset += 4

                            if value != -2147483648:
                                self.resp[self.dp_sym[topic_id]][
                                    self.depthvalue[index]
                                ] = value

                        offset += 2

                        multiplier = struct.unpack(">H", data[offset : offset + 2])[0]
                        self.resp[self.dp_sym[topic_id]]["multiplier"] = multiplier
                        offset += 2
                        precision = struct.unpack("B", data[offset : offset + 1])[0]
                        self.resp[self.dp_sym[topic_id]]["precision"] = precision
                        offset += 1

                        val = ["exchange", "exchange_token", "symbol"]
                        for i in range(3):
                            string_len = struct.unpack("B", data[offset : offset + 1])[
                                0
                            ]
                            offset += 1
                            string_data = data[offset : offset + string_len].decode(
                                "utf-8",errors='ignore'
                            )
                            self.resp[self.dp_sym[topic_id]][val[i]] = string_data
                            offset += string_len
                        self.resp[self.dp_sym[topic_id]]["type"] = "dp"
                        self.resp[topic_name]["symbol"] = self.symbol_token[topic_name]
                        self.__response_output(
                            self.resp[self.dp_sym[topic_id]], "depth"
                        )

                    elif topic_name[:2] == "if":

                        self.index_sym[topic_id] = topic_name
                        self.resp[self.index_sym[topic_id]] = {}

                        # field_count - 21 in scrips , 25 in depth , 6 in index
                        field_count = struct.unpack("B", data[offset : offset + 1])[0]
                        offset += 1

                        for index in range(field_count):

                            value = struct.unpack(">i", data[offset : offset + 4])[0]
                            offset += 4

                            if value != -2147483648:
                                self.resp[self.index_sym[topic_id]][
                                    self.index_val[index]
                                ] = value

                        offset += 2

                        multiplier = struct.unpack(">H", data[offset : offset + 2])[0]
                        self.resp[self.index_sym[topic_id]]["multiplier"] = multiplier
                        offset += 2

                        precision = struct.unpack("B", data[offset : offset + 1])[0]
                        self.resp[self.index_sym[topic_id]]["precision"] = precision
                        offset += 1

                        val = ["exchange", "exchange_token", "symbol"]
                        for i in range(3):
                            string_len = struct.unpack("B", data[offset : offset + 1])[
                                0
                            ]
                            offset += 1
                            string_data = data[offset : offset + string_len].decode(
                                "utf-8",errors='ignore'
                            )
                            self.resp[self.index_sym[topic_id]][val[i]] = string_data
                            offset += string_len
                        self.resp[topic_name]["symbol"] = self.symbol_token[topic_name]
                        self.resp[self.index_sym[topic_id]]["type"] = "if"
                        self.__response_output(
                            self.resp[self.index_sym[topic_id]], "index"
                        )

                    elif topic_name[:2] == "sf":
                        self.scrips_sym[topic_id] = topic_name
                        self.resp[self.scrips_sym[topic_id]] = {}

                        # field_count - 21 in scrips , 25 in depth , 6 in index
                        field_count = struct.unpack("B", data[offset : offset + 1])[0]
                        offset += 1

                        for index in range(field_count):
                            value = struct.unpack(">i", data[offset : offset + 4])[0]
                            offset += 4
                            if value != -2147483648:
                                self.resp[self.scrips_sym[topic_id]][
                                    self.data_val[index]
                                ] = value

                        offset += 2

                        multiplier = struct.unpack(">H", data[offset : offset + 2])[0]
                        self.resp[self.scrips_sym[topic_id]]["multiplier"] = multiplier
                        offset += 2

                        precision = struct.unpack("B", data[offset : offset + 1])[0]
                        self.resp[self.scrips_sym[topic_id]]["precision"] = precision
                        offset += 1
                        val = ["exchange", "exchange_token", "symbol"]
                        for i in range(3):
                            string_len = struct.unpack("B", data[offset : offset + 1])[
                                0
                            ]
                            offset += 1
                            string_data = bytes(data[offset : offset + string_len]).decode(
                                "utf-8" ,errors='ignore'
                            )
                            self.resp[self.scrips_sym[topic_id]][val[i]] = string_data
                            offset += string_len
                        self.resp[topic_name]["symbol"] = self.symbol_token[topic_name]
                        self.resp[self.scrips_sym[topic_id]]["type"] = "sf"
                        self.__response_output(
                            self.resp[self.scrips_sym[topic_id]], "scrips"
                        )

                elif data_type == 85:  # Full mode datafeed
                    offset += 1
                    topic_id = struct.unpack("H", data[offset : offset + 2])[0]
                    offset += 2

                    field_count = struct.unpack("B", data[offset : offset + 1])[0]
                    offset += 1
                    sf_flag, idx_flag, dp_flag = False, False, False
                    self.UpdateTick = False
                    for index in range(field_count):
                        value = struct.unpack(">i", data[offset : offset + 4])[0]
                        offset += 4
                        # if field_count == 20 or field_count == 21:
                        if topic_id in self.scrips_sym:
                            if self.data_val[index] in self.resp[self.scrips_sym[topic_id]] and self.resp[self.scrips_sym[topic_id]][
                                self.data_val[index]] != value and value != -2147483648:
                                self.resp[self.scrips_sym[topic_id]][
                                    self.data_val[index]
                                ] = value
                                self.UpdateTick = True
                            elif self.data_val[index] not in self.resp[self.scrips_sym[topic_id]] and value != -2147483648:
                                self.resp[self.scrips_sym[topic_id]][
                                    self.data_val[index]
                                ] = value
                                self.UpdateTick = True

                            sf_flag = True
                        elif topic_id in self.index_sym:
                            if self.index_val[index] in self.resp[self.index_sym[topic_id]] and  self.resp[self.index_sym[topic_id]][self.index_val[index]] != value and value != "-2147483648":

                                self.resp[self.index_sym[topic_id]][
                                    self.index_val[index]
                                ] = value
                                self.UpdateTick = True
                            elif self.index_val[index] not in self.resp[self.index_sym[topic_id]] and value != -2147483648:
                                self.resp[self.index_sym[topic_id]][
                                    self.index_val[index]
                                ] = value
                                self.UpdateTick = True
                            idx_flag = True
                        elif topic_id in self.dp_sym:
                            if self.depthvalue[index] in self.resp[self.dp_sym[topic_id]] and self.resp[self.dp_sym[topic_id]][
                                self.depthvalue[index]] != value and value != -2147483648:
                                self.resp[self.dp_sym[topic_id]][
                                    self.depthvalue[index]
                                ] = value
                                self.UpdateTick = True
                            elif self.depthvalue[index] not in self.resp[self.dp_sym[topic_id]] and  value != -2147483648:
                                self.resp[self.dp_sym[topic_id]][
                                    self.depthvalue[index]
                                ] = value
                                self.UpdateTick = True                                
                            dp_flag = True
                    if self.UpdateTick:
                        if sf_flag:
                            self.__response_output(
                                self.resp[self.scrips_sym[topic_id]], "scrips"
                            )
                        elif idx_flag:
                            self.__response_output(
                                self.resp[self.index_sym[topic_id]], "index"
                            )
                        elif dp_flag:
                            self.__response_output(
                                self.resp[self.dp_sym[topic_id]], "depth"
                            )

                elif data_type == 76:  # lite mode datafeed

                    offset += 1
                    topic_id = struct.unpack("H", data[offset : offset + 2])[0]
                    offset += 2
                    sf_flag, idx_flag = False, False
                    if topic_id in self.scrips_sym:

                        # for index in range(3):
                        value = struct.unpack(">i", data[offset : offset + 4])[0]
                        offset += 4
                        if value != self.resp[self.scrips_sym[topic_id]][self.data_val[0]] and value != -2147483648:
                            self.resp[self.scrips_sym[topic_id]][self.data_val[0]] = value
                            sf_flag = True
                            self.resp[self.scrips_sym[topic_id]]["type"] = "sf"
                            self.__response_output(
                                self.resp[self.scrips_sym[topic_id]], "scrips"
                            )
                    elif topic_id in self.index_sym:
                        value = struct.unpack(">i", data[offset : offset + 4])[0]
                        offset += 4
                        if value != self.resp[self.index_sym[topic_id]][self.index_val[0]] and value != -2147483648:
                            self.resp[self.index_sym[topic_id]][self.index_val[0]] = value
                            idx_flag = True
                            self.resp[self.index_sym[topic_id]]["type"] = "if"
                        
                            self.__response_output(
                                self.resp[self.index_sym[topic_id]], "index"
                            )
                        
        except Exception as e:
            self.data_logger.exception(e)


    def __response_msg(self, data: bytearray):
        """
        Processes the response message based on the response type and calls the corresponding function.

        Args:
            data (bytearray): The response data.

        """
        try:

            _, resp_type = struct.unpack("!HB", data[:3])
            if resp_type == 1:  # Authentication response
                self.__auth_resp(data)

            elif resp_type == 4:  # subsciption response
                self.__subscribe_resp(data)

            elif resp_type == 5:  # Unsubsciption response
                self.__unsubscribe_resp(data)

            elif resp_type == 6:  # Data Feed Response
                self.__datafeed_resp(data)

            elif resp_type == 7 or resp_type == 8:
                self.__resume_pause_resp(data, resp_type)

            elif resp_type == 12:  # Full Mode Data Response
                self.__lite_full_mode_resp(data)

        except Exception as e:
            self.data_logger.exception(e)

    def __symbol_conversion(self, symbolslst: list) -> dict:
        """
        Converts symbols to HSM symbol tokens and returns a dictionary of {hsmtoken: symbol}.

        Args:
            symbolslst (list): A list of symbols to convert.

        Returns:
            dict: A dictionary mapping HSM symbol tokens to symbols.
        """
        try:

            wrong_symbols = []
            symb_flag = False
            idx_dp_flag = False
            symbol_dict = {}
            total_symbols = len(symbolslst)
            if (
                len(self.scrips_per_channel[self.channel_num]) > self.symbol_limit
                or total_symbols > self.symbol_limit
                and len(self.scrips_per_channel[self.channel_num]) + total_symbols > self.symbol_limit
            ):
                self.On_error(
                    {
                        "code": defines.LIMIT_EXCEED_CODE,
                        "message": defines.LIMIT_EXCEED_MSG_5000,
                        "s": defines.ERROR,
                    }
                )
                return 

            symbol_chunks = [
                symbolslst[i : i + 500] for i in range(0, total_symbols, 500)
            ]
            conv = SymbolConversion(self.__access_token, self.data_type, self.log_path)
            for symbols in symbol_chunks:
                symbol_value = conv.symbol_to_hsmtoken(symbols)
                if symbol_value[3] != "":
                    self.On_error(
                    {
                        "code": defines.INVALID_CODE,
                        "message": symbol_value[3],
                        "s": defines.ERROR,
                        "type": defines.SUBS_TYPE
                    })
                    return
                symbol_dict.update(symbol_dict, **symbol_value[0])
                if type(symbol_value[1]) == list and len(symbol_value[1]) > 0:

                    wrong_symbols += symbol_value[1]
                    symb_flag = True
                if symbol_value[2] == True:
                    idx_dp_flag = True
            
                
            if symb_flag:
                self.On_error(
                    {
                        "code": defines.INVALID_CODE,
                        "message": defines.INVALID_SYMBOLS,
                        "s": defines.ERROR,
                        "type": defines.SUBS_TYPE,
                        "invalid_symbols": wrong_symbols,
                    }
                )
            if idx_dp_flag:
                self.On_error(
                    {
                        "code": defines.INVALID_CODE,
                        "message": defines.INDEX_DEPTH_ERROR_MESSAGE,
                        "s": defines.ERROR,
                        "type": defines.SUBS_TYPE,
                    }
                )

            return symbol_dict

        except Exception as e:
            self.data_logger.exception(e)

    def __channel_resume_pause(self):
        """
        Pauses the active channel and resumes the specified channel if necessary.

        If the WebSocket object (__ws_object) is not None and there is an active channel (active_channel)
        that is different from the specified channel (channelNum), the function creates and appends a pause message
        for the active channel to the message list. If the specified channel is already in the running_channels set,
        the function creates and appends a resume message for the specified channel to the message list.

        Finally, it updates the running_channels set and sets the active_channel to the specified channel (channelNum).

        """
        try:
            if (
                self.__ws_object is not None
                and self.active_channel is not None
                and self.active_channel != self.channel_num
            ):
                message = self.__channel_pause_msg(self.active_channel)
                # self.message.append(message)
                self.add_message(message)


                if self.channel_num in self.running_channels:
                    message = self.__channel_resume_msg(self.channel_num)
                    # self.message.append(message)
                    self.add_message(message)

            self.running_channels.add(self.channel_num)
            self.active_channel = self.channel_num

        except Exception as e:
            self.data_logger.exception(e)

    def channel_resume(self, channel: int) -> None:
        """
        Resumes the specified channel.

        Args:
            channel (int): The channel number to resume.
        """
        try:
            self.channel_num = channel
            self.__channel_resume_pause()

        except Exception as e:
            self.data_logger.exception(e)

    def On_message(self, message: dict) -> None:
        """
        Handles the received message.

        Args:
            message (str): The received message.
        """
        try:
            if self.OnMessage is not None:
                self.OnMessage(message)
            else:
                if self.write_to_file:
                    self.data_logger.debug(f"Response:{message}")
                else:
                    print(f"Response:{message}")

        except KeyError as e:
            key_name = str(e)
            self.data_logger.exception(e)
            self.On_error(f"KeyError: The key {key_name} is missing in the response.")

        except Exception as e:
            self.data_logger.exception(e)
            self.On_error(e)

    def __send_message(self, message: str) -> None:
        """
        Sends a message through the WebSocket connection.

        Args:
            message (str): The message to send.
        """
        with self.websocket_lock:
            if self.__ws_object is not None:

                self.__ws_object.send(message, opcode=websocket.ABNF.OPCODE_BINARY)


    def add_message(self, message):
        """
        Add a message to the list of messages and notify waiting threads.

        Args:
            message (str): The message to add to the list.
        """
        with self.message_lock:
            self.message.append(message)
            self.message_condition.notify()


    def __process_message_queue(self) -> None:
        """
        Processes the message queue by sending messages sequentially.
        """

        while not self.message_thread_stop_event.is_set():
            with self.message_lock:
                while not self.message_thread_stop_event.is_set() and  not self.message:  # Use a loop to handle spurious wake-ups
                    self.message_condition.wait()
                if self.message_thread_stop_event.is_set():
                    break
                # Once a message is available, pop it from the queue
                message = self.message.pop(0)

            # Send the message outside the lock to avoid blocking other threads
            self.__send_message(message)
            # time.sleep(0.01)


    def On_error(self, message: dict) -> None:
        """
        Handles the error message.

        Args:
            message (str): The error message.
        """
        if self.OnError is not None:
            self.OnError(message)
            self.data_logger.error(message)
        else:
            if self.write_to_file:
                self.data_logger.debug(f"ERROR Response:{message}")
            else:
                print(f"Error: {message}")




    def on_open(self) -> None:
        """
        Handles the open action.
        """
        try:
            if self.OnOpen:
                self.OnOpen()
        except Exception as e:
            self.data_logger.exception(e)
            self.On_error(e)


    def connect(self) -> None:
        """
        Establishes a connection to the WebSocket.

        If the WebSocket object is not already initialized, this method will create the
        WebSocket connection.

        """
        try:
            if self.__ws_object is None:
                self.__init_connection()
                time.sleep(2)
            self.on_open()

        except Exception as e:
            self.data_logger.exception(e)
            self.On_error(e)
            

    def on_close(self, message: dict) -> None:
        """
        Handles the close event.

        Args:
            message (dict): The close message .
        """
        try:
            if self.OnClose:
                self.OnClose(message)
            else:
                print(f"Response: {message}")
        except Exception as e:
            self.data_logger.exception(e)
            self.On_error(e)


    def __on_open(self, ws) -> None:
        """
        Handles the WebSocket connection open event.

        Args:
            ws (websocket.WebSocketApp): The WebSocket object.
        """
        if self.__ws_object is None:
            self.message = []
            self.__ws_object = ws
            self.message_thread = Thread(target=self.__process_message_queue)
            self.ping_thread = Thread(target=self.__ping)
            self.message_thread_stop_event = threading.Event()  # Event to signal stopping
            self.message_thread.start()
            self.ping_thread.start()
            message = self.__access_token_msg()
            # self.message.append(message)
            self.add_message(message)
            self.reconnect_attempts = 0
            self.reconnect_delay = 0
            if self.lite:
                message = self.__lite_mode_msg()
                # self.message.append(message)
                self.add_message(message)
            else:
                message = self.__full_mode_msg()
                # self.message.append(message)
                self.add_message(message)


    def __on_close(self, ws, close_code, close_reason):
        """
        Handle the WebSocket connection close event.

        Args:
            ws (WebSocket): The WebSocket object.
            close_code (int): The code indicating the reason for closure.
            close_reason (str): The reason for closure.

        Returns:
            dict: A dictionary containing the response code, message, and s.
        """
        if self.restart_flag:
            if self.reconnect_attempts < self.max_reconnect_attempts:

                if self.write_to_file:
                    self.data_logger.debug(
                        f"Response:{f'Attempting reconnect {self.reconnect_attempts+1} of {self.max_reconnect_attempts}...'}"
                    )
                else:
                    print(
                        f"Attempting reconnect {self.reconnect_attempts+1} of {self.max_reconnect_attempts}..."
                    )

                if (self.reconnect_attempts) % 5 == 0:
                    self.reconnect_delay += 5
                time.sleep(self.reconnect_delay)
                self.reconnect_attempts += 1

                self.__ws_object = None
                self.scrips_per_channel[self.channel_num] = []
                self.symbol_token = {}

                self.connect()
            else:
                if self.write_to_file:
                    self.data_logger.debug(
                        f"Response:{'Max reconnect attempts reached. Connection abandoned.'}"
                    )
                else:
                    print("Max reconnect attempts reached. Connection abandoned.")
        else:

            self.on_close(
                {
                    "code": defines.SUCCESS_CODE,
                    "message": defines.CONNECTION_CLOSED,
                    "s": defines.SUCCESS,
                }
            )

    def __ping(self):
        while (
            self.__ws_object is not None
            and self.__ws_object.sock
            and self.__ws_object.sock.connected
        ):
            self.__ws_object.send(bytes([0, 1, 11]), opcode=websocket.ABNF.OPCODE_BINARY)
            time.sleep(10)

    def __init_connection(self):
        """
        Initializes the WebSocket connection and starts the WebSocketApp.

        The method creates a WebSocketApp object with the specified URL and sets the appropriate event handlers.
        It then starts the WebSocketApp in a separate thread.
        """
        try:
            if  self.access_token_to_hsmtoken() and self.__valid_token :
                if self.write_to_file:
                    self.background_flag = True  
                ws = websocket.WebSocketApp(
                    self.__url,
                    on_message=lambda ws, msg: self.__response_msg(msg),
                    on_error=lambda ws, msg: self.On_error(msg),
                    on_close=lambda ws, close_code, close_reason: self.__on_close(
                        ws, close_code, close_reason
                    ),
                    on_open=lambda ws: self.__on_open(ws),
                )

                self.ws_thread = Thread(target=ws.run_forever)
                self.ws_thread.daemon = self.background_flag
                self.ws_thread.start()

        except Exception as e:
            self.data_logger.exception(e)

    def close_connection(self) -> None:
        """
        Closes the WebSocket connection 

        """

        if self.__ws_object:
            self.restart_flag = False
            self.__ws_object.close()
            self.__ws_object = None
            self.ws_thread.join()
            self.message_thread_stop_event.set()
            with self.message_lock:
                self.message_condition.notify()  # Notify the thread to wake up
            self.message_thread.join()
            self.ping_thread.join()
            self.__ws_run = False
            self.scrips_per_channel[self.channel_num] = []


    def keep_running(self):
        """
        Starts an infinite loop to keep the program running.

        """
        self.__ws_run = True
        self.infy_loop = Thread(target=self.infinite_loop)
        self.infy_loop.start()

    def infinite_loop(self):
        while self.__ws_run:
            time.sleep(0.5)


    def is_connected(self):
        """
        Check if the websocket is connected.

        Returns:
            bool: True if the websocket is connected, False otherwise.
        """
        if self.__ws_object:
            return True
        else:
            return False

    def unsubscribe(
        self, symbols: list, data_type: str = "SymbolUpdate", channel: int = 11
    ):
        """
        Unsubscribes from real-time data updates for the specified symbols.

        Args:
            symbols (list): A list of symbols to unsubscribe from.
            data_type (str, optional): The type of data to unsubscribe from. Defaults to "SymbolUpdate".
            channel (int, optional): The channel to use for unsubscription. Defaults to 1.
        """
        try:
            if self.__valid_token:
                self.data_type = data_type
                self.symbols = symbols
                self.channel_num = channel
                self.__channel_resume_pause()
                self.channel_symbol = self.__symbol_conversion(symbols)
                self.unsub_symbol = list(self.channel_symbol.keys())
                for symb in self.unsub_symbol:
                    if symb not in self.scrips_count[self.channel_num]:
                        self.unsub_symbol.remove(symb)

                if len(self.unsub_symbol) != 0:
                    total_symbols = len(self.unsub_symbol)
                    symbol_chunks = [
                        self.unsub_symbol[i : i + 1500] for i in range(0, total_symbols, 1500)
                    ]
                    for symbols in symbol_chunks:
                        message = self.__unsubscription_msg(symbols)
                        # self.message.append(message)
                        self.add_message(message)

                else:
                    self.On_error(
                        {
                            "code": defines.INVALID_CODE,
                            "message": defines.INVALID_SYMBOLS,
                            "s": defines.ERROR,
                        }
                    )

        except Exception as e:
            self.data_logger.exception(e)

    def subscribe(
        self, symbols: list, data_type: str = "SymbolUpdate", channel: int = 11
    ):
        """
        Subscribes to real-time data updates for the specified symbols.

        Args:
            symbols (list): A list of symbols to subscribe to.
            data_type (str, optional): The type of data to subscribe to. Defaults to "SymbolUpdate".
            channel (int, optional): The channel to use for subscription. Defaults to 1.
        """
        try:
            if self.__valid_token:
                self.data_type = data_type
                self.symbols = symbols
                self.channel_num = channel
                self.__channel_resume_pause()
                self.channel_symbol = self.__symbol_conversion(symbols)
                if self.channel_symbol is None:
                    return
                if len(self.symbol_token) + len(self.channel_symbol) > 5000:
                    self.On_error(
                        {
                            "code": defines.LIMIT_EXCEED_CODE,
                            "message": defines.LIMIT_EXCEED_MSG_5000,
                            "s": defines.ERROR,
                        }
                    )
                    return
                self.symbol_token.update(self.symbol_token, **self.channel_symbol)
                self.scrips_count[self.channel_num] = list(self.channel_symbol.keys())
                total_symbols = len(self.scrips_count[self.channel_num])
                symbol_chunks = [
                    self.scrips_count[self.channel_num][i : i + 1500]
                    for i in range(0, total_symbols, 1500)
                ]
                for symbols in symbol_chunks:
                    message = self.__subscription_msg(symbols)
                    # self.message.append(message)
                    time.sleep(0.5)
                    self.add_message(message)


        except Exception as e:
            self.data_logger.exception(e)



import requests
import time
import traceback
import datetime


class UpstoxHist():
    LTPData_per_sec = 1
    MaxQuoteSymbolLimit = 500
    _timeout = 15

    BASE_URL = "https://api.upstox.com/v2/"
    ENDPOINTS = { # For V3
        "getData" : "historical-candle/{instrument_key}/{interval}/{to_date}/{from_date}",
        "getDataIntraday": "historical-candle/intraday/{instrument_key}/{interval}",
        "getQuotes" : "market-quote/quotes"
    }

    HistoricalMapping = {
        "1": ['1minute', 28],    # Upstox has 28 days of 1-minute data
        "30": ['30minute', 364],
        '1D': ['day', 364],
        '1W': ['week', 3640],
        '1M': ['month', 3640]
    }
    headers = {'Accept': 'application/json',
               'Content-Type': 'application/json'}


    def __init__(self,sessionData : dict):
        self.access_token = sessionData['Sessionid']
        self.reqsession = requests.Session()
        self.__supportedFormats = list(self.info()['supported timeformats']["timeValues"].values())
        self.set_accesstoken(self.access_token)

    def set_accesstoken(self,accesstoken):
        self.headers['Authorization'] = f'Bearer {accesstoken}'

    def requestAPI(self, method, url, body=None,params=None, timeout=_timeout, headers = None):
        headers = self.headers if headers == None else headers
        try:
            resp = self.reqsession.request(method, url, headers=headers, json=body, timeout=timeout, params=params)
            data = resp.json()
            if resp.status_code == 200 :
                if data.get("status") == "success" :
                    if isinstance(data['data'], list):
                        return {"status": True, "data": data['data'], "error": False, "message": "Data Received"}
                    else:
                        return {"status": True, "data": [data['data']], "error": False, "message": "Data Received"}
                else:
                    return {'status' : True, "data" : [data], "error": False, "message" : "Data Received"}

            else:
                try :
                    return {"status": False, "data": [], "error": True, "message": data['errors'][0]['message']}
                except :
                    return  {"status": False, "data": [], "error": True, "message": data}

        except requests.exceptions.Timeout:
            return {"status": False, "data": [], "error": True, "message": "Timeout error"}
        except Exception as e:
            try:
                return {"status": False, "data": [], "error": True, "message": resp.text}
            except:
                return {"status": False, "data": [], "error": True, "message":  "Error - Unable to receive data."}

    def get_quotes(self, symbols: list,sleeptime = 0):
        main_data = {}
        tokenMap = {}
        nonTokens = []

        # Map tokens to exchange:symbol and categorize tokens
        for token in symbols:
            histToken = token.get('histToken')
            if histToken:
                tokenMap[histToken] = token['token']
            else:
                nonTokens.append(token['token'])

        if not tokenMap:
            return {
                "status": False,
                "data": [],
                "error": True,
                "message": "No histToken found for any symbols"
            }

        tokenMapKeys = list(tokenMap.keys())
        for i in range(0, len(tokenMapKeys), self.MaxQuoteSymbolLimit):
            instrument_keys = tokenMapKeys[i:i + self.MaxQuoteSymbolLimit]
            query_params = {
                "instrument_key": ",".join(instrument_keys),
            }
            url = f'{self.BASE_URL}{self.ENDPOINTS["getQuotes"]}'
            try:
                response = self.requestAPI("GET", url, params=query_params)
                if not response.get('status') or response.get('error'):
                    return response

                api_data = response.get('data', [])
                if not api_data or not api_data[0]:
                    return {
                        "status": False,
                        "data": [],
                        "error": True,
                        "message": "No data received"
                    }
                api_data = api_data[0]
                for key , message in api_data.items():
                    main_data[tokenMap[message['instrument_token']]] = {
                        "timestamp_str": datetime.datetime.fromtimestamp(int(message['last_trade_time'])//1000).strftime("%Y-%m-%d %H:%M:%S"),
                        "timestamp": str(int(message['last_trade_time'])//1000),
                        "symbol": message['instrument_token'],
                        "ltp": message['last_price'],
                        "prev_day_close": message['ohlc']['close'],
                        "oi": message['oi'],
                        "prev_day_oi": message['oi'],
                        "turnover": message['volume'],
                        "best_bid_price": message['depth']['buy'][0]['price'] if message['depth']['buy'] else 0,
                        "best_ask_price": message['depth']['sell'][0]['price'] if message['depth']['sell'] else 0,
                        "best_bid_qty": message['depth']['buy'][0]['quantity'] if message['depth']['buy'] else 0,
                        "best_ask_qty": message['depth']['sell'][0]['quantity'] if message['depth']['sell'] else 0,
                        "ttq": message['volume'],
                        "token": message['instrument_token']
                    }
            except Exception as e:
                return {
                    "status": False,
                    "data": [],
                    "error": True,
                    "message": f"Error occurred: {str(e)}"
                }
            time.sleep(sleeptime)

        # Fill in empty data for nonTokens
        for token in nonTokens:
            main_data[token] = {}

        return {
            "status": True,
            "data": main_data,
            "error": False,
            "message": "Data Received"
        }

    def getHistorical(self, token, startTime : datetime, endTime : datetime, interval : str): #,sec = False
        if interval not in self.__supportedFormats:
            return {"status" : False, "error" : True, "message" : f"Invalid interval. Please input as per the format: {self.info()}", "data" : []}

        if not token:
            return {"status" : False, "data" : [], "error" : True, "message" : "No histToken found for the symbol"}

        main_data = []
        # check if today's date is in range of start and end date
        if datetime.datetime.now().date() >= startTime.date() and datetime.datetime.now().date() <= endTime.date()  and interval in ['1', '30']:
            Intradays = self.getIntradays(token, self.HistoricalMapping[interval][0])
            if Intradays.get('status') and not Intradays.get('error'):
                main_data.extend(Intradays.get('data', []))
            else:
                return Intradays

        endTime = min(endTime, datetime.datetime.now())
        while startTime < endTime:
            ed = min(startTime + datetime.timedelta(days=self.HistoricalMapping[interval][1]-1), endTime)

            url = f'{self.BASE_URL}{self.ENDPOINTS["getData"]}'.format(
                instrument_key=token,
                interval=self.HistoricalMapping[interval][0],
                to_date=ed.strftime("%Y-%m-%d"),
                from_date=startTime.strftime("%Y-%m-%d")
            )
            response = self.requestAPI("GET", url)
            if not response.get('status') or response.get('error'):
                return response

            candle_data = response.get('data', [])

            if candle_data:
                main_data.extend(self.ConvertToSTD(candle_data[0].get('candles', [])))
            else:
                return {"status": False, "data": [], "error": True, "message": "Data not available"}

            startTime = ed

        return {"status": True, "data": main_data, "error": False, "message": "Data Received"} if main_data else \
                {"status": False, "data": [], "error": True, "message": "Data not available"}

    def ConvertToSTD(self, data):
        if not data:
            return []

        std_format = [
            {
                'datetime': datetime.datetime.fromisoformat(candle[0]).strftime("%Y-%m-%d %H:%M:%S"),
                'open': float(candle[1]),
                'high': float(candle[2]),
                'low': float(candle[3]),
                'close': float(candle[4]),
                'volume': int(candle[5]),
                'oi': int(candle[6]) if len(candle) == 7 else 0
            }
            for candle in data
        ]

        return std_format

    def getIntradays(self, token, interval : str):
        url = f'{self.BASE_URL}{self.ENDPOINTS["getDataIntraday"]}'.format(
            instrument_key=token,
            interval=interval
        )
        try:
            response = self.requestAPI("GET", url)
            if not response.get('status') or response.get('error'):
                return response

            candle_data = response.get('data', [])

            if candle_data:
                main_data = self.ConvertToSTD(candle_data[0].get('candles', []))
                return {"status": True, "data": main_data, "error": False, "message": "Data Received"}
            else:
                return {"status": False, "data": [], "error": True, "message": "Data not available"}
        except Exception as e:
            return {
                "status": False,
                "data": [],
                "error": True,
                "message": f"Error occurred: {str(e)}"
            }

    def info(self):
        return {"supported timeformats" :
            {"timeValues" :
                {
                    "1 minute" : "1",
                    "30 minute" : "30",
                    "1 Day" : "1D",
                    "1 Week" : "1W",
                    "1 Month" : "1M"
                }}}




# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: MarketDataFeed.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14MarketDataFeed.proto\x12%com.upstox.marketdatafeeder.rpc.proto\"9\n\x04LTPC\x12\x0b\n\x03ltp\x18\x01 \x01(\x01\x12\x0b\n\x03ltt\x18\x02 \x01(\x03\x12\x0b\n\x03ltq\x18\x03 \x01(\x03\x12\n\n\x02\x63p\x18\x04 \x01(\x01\"P\n\x0bMarketLevel\x12\x41\n\x0b\x62idAskQuote\x18\x01 \x03(\x0b\x32,.com.upstox.marketdatafeeder.rpc.proto.Quote\"G\n\nMarketOHLC\x12\x39\n\x04ohlc\x18\x01 \x03(\x0b\x32+.com.upstox.marketdatafeeder.rpc.proto.OHLC\"Q\n\x05Quote\x12\n\n\x02\x62q\x18\x01 \x01(\x05\x12\n\n\x02\x62p\x18\x02 \x01(\x01\x12\x0b\n\x03\x62no\x18\x03 \x01(\x05\x12\n\n\x02\x61q\x18\x04 \x01(\x05\x12\n\n\x02\x61p\x18\x05 \x01(\x01\x12\x0b\n\x03\x61no\x18\x06 \x01(\x05\"z\n\x0cOptionGreeks\x12\n\n\x02op\x18\x01 \x01(\x01\x12\n\n\x02up\x18\x02 \x01(\x01\x12\n\n\x02iv\x18\x03 \x01(\x01\x12\r\n\x05\x64\x65lta\x18\x04 \x01(\x01\x12\r\n\x05theta\x18\x05 \x01(\x01\x12\r\n\x05gamma\x18\x06 \x01(\x01\x12\x0c\n\x04vega\x18\x07 \x01(\x01\x12\x0b\n\x03rho\x18\x08 \x01(\x01\"\xbf\x02\n\x13\x45xtendedFeedDetails\x12\x0b\n\x03\x61tp\x18\x01 \x01(\x01\x12\n\n\x02\x63p\x18\x02 \x01(\x01\x12\x0b\n\x03vtt\x18\x03 \x01(\x03\x12\n\n\x02oi\x18\x04 \x01(\x01\x12\x10\n\x08\x63hangeOi\x18\x05 \x01(\x01\x12\x11\n\tlastClose\x18\x06 \x01(\x01\x12\x0b\n\x03tbq\x18\x07 \x01(\x01\x12\x0b\n\x03tsq\x18\x08 \x01(\x01\x12\r\n\x05\x63lose\x18\t \x01(\x01\x12\n\n\x02lc\x18\n \x01(\x01\x12\n\n\x02uc\x18\x0b \x01(\x01\x12\n\n\x02yh\x18\x0c \x01(\x01\x12\n\n\x02yl\x18\r \x01(\x01\x12\n\n\x02\x66p\x18\x0e \x01(\x01\x12\n\n\x02\x66v\x18\x0f \x01(\x05\x12\x0e\n\x06mbpBuy\x18\x10 \x01(\x03\x12\x0f\n\x07mbpSell\x18\x11 \x01(\x03\x12\n\n\x02tv\x18\x12 \x01(\x03\x12\x0c\n\x04\x64hoi\x18\x13 \x01(\x01\x12\x0c\n\x04\x64loi\x18\x14 \x01(\x01\x12\n\n\x02sp\x18\x15 \x01(\x01\x12\x0b\n\x03poi\x18\x16 \x01(\x01\"l\n\x04OHLC\x12\x10\n\x08interval\x18\x01 \x01(\t\x12\x0c\n\x04open\x18\x02 \x01(\x01\x12\x0c\n\x04high\x18\x03 \x01(\x01\x12\x0b\n\x03low\x18\x04 \x01(\x01\x12\r\n\x05\x63lose\x18\x05 \x01(\x01\x12\x0e\n\x06volume\x18\x06 \x01(\x05\x12\n\n\x02ts\x18\x07 \x01(\x03\"\xf8\x02\n\x0eMarketFullFeed\x12\x39\n\x04ltpc\x18\x01 \x01(\x0b\x32+.com.upstox.marketdatafeeder.rpc.proto.LTPC\x12G\n\x0bmarketLevel\x18\x02 \x01(\x0b\x32\x32.com.upstox.marketdatafeeder.rpc.proto.MarketLevel\x12I\n\x0coptionGreeks\x18\x03 \x01(\x0b\x32\x33.com.upstox.marketdatafeeder.rpc.proto.OptionGreeks\x12\x45\n\nmarketOHLC\x18\x04 \x01(\x0b\x32\x31.com.upstox.marketdatafeeder.rpc.proto.MarketOHLC\x12P\n\x0c\x65\x46\x65\x65\x64\x44\x65tails\x18\x05 \x01(\x0b\x32:.com.upstox.marketdatafeeder.rpc.proto.ExtendedFeedDetails\"\xbc\x01\n\rIndexFullFeed\x12\x39\n\x04ltpc\x18\x01 \x01(\x0b\x32+.com.upstox.marketdatafeeder.rpc.proto.LTPC\x12\x45\n\nmarketOHLC\x18\x02 \x01(\x0b\x32\x31.com.upstox.marketdatafeeder.rpc.proto.MarketOHLC\x12\x11\n\tlastClose\x18\x03 \x01(\x01\x12\n\n\x02yh\x18\x04 \x01(\x01\x12\n\n\x02yl\x18\x05 \x01(\x01\"\xaf\x01\n\x08\x46ullFeed\x12I\n\x08marketFF\x18\x01 \x01(\x0b\x32\x35.com.upstox.marketdatafeeder.rpc.proto.MarketFullFeedH\x00\x12G\n\x07indexFF\x18\x02 \x01(\x0b\x32\x34.com.upstox.marketdatafeeder.rpc.proto.IndexFullFeedH\x00\x42\x0f\n\rFullFeedUnion\"\xa8\x02\n\x0bOptionChain\x12\x39\n\x04ltpc\x18\x01 \x01(\x0b\x32+.com.upstox.marketdatafeeder.rpc.proto.LTPC\x12\x41\n\x0b\x62idAskQuote\x18\x02 \x01(\x0b\x32,.com.upstox.marketdatafeeder.rpc.proto.Quote\x12I\n\x0coptionGreeks\x18\x03 \x01(\x0b\x32\x33.com.upstox.marketdatafeeder.rpc.proto.OptionGreeks\x12P\n\x0c\x65\x46\x65\x65\x64\x44\x65tails\x18\x04 \x01(\x0b\x32:.com.upstox.marketdatafeeder.rpc.proto.ExtendedFeedDetails\"\xd1\x01\n\x04\x46\x65\x65\x64\x12;\n\x04ltpc\x18\x01 \x01(\x0b\x32+.com.upstox.marketdatafeeder.rpc.proto.LTPCH\x00\x12=\n\x02\x66\x66\x18\x02 \x01(\x0b\x32/.com.upstox.marketdatafeeder.rpc.proto.FullFeedH\x00\x12@\n\x02oc\x18\x03 \x01(\x0b\x32\x32.com.upstox.marketdatafeeder.rpc.proto.OptionChainH\x00\x42\x0b\n\tFeedUnion\"\xf3\x01\n\x0c\x46\x65\x65\x64Response\x12\x39\n\x04type\x18\x01 \x01(\x0e\x32+.com.upstox.marketdatafeeder.rpc.proto.Type\x12M\n\x05\x66\x65\x65\x64s\x18\x02 \x03(\x0b\x32>.com.upstox.marketdatafeeder.rpc.proto.FeedResponse.FeedsEntry\x1aY\n\nFeedsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12:\n\x05value\x18\x02 \x01(\x0b\x32+.com.upstox.marketdatafeeder.rpc.proto.Feed:\x02\x38\x01*\'\n\x04Type\x12\x10\n\x0cinitial_feed\x10\x00\x12\r\n\tlive_feed\x10\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'MarketDataFeed_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _FEEDRESPONSE_FEEDSENTRY._options = None
  _FEEDRESPONSE_FEEDSENTRY._serialized_options = b'8\001'
  _TYPE._serialized_start=2421
  _TYPE._serialized_end=2460
  _LTPC._serialized_start=63
  _LTPC._serialized_end=120
  _MARKETLEVEL._serialized_start=122
  _MARKETLEVEL._serialized_end=202
  _MARKETOHLC._serialized_start=204
  _MARKETOHLC._serialized_end=275
  _QUOTE._serialized_start=277
  _QUOTE._serialized_end=358
  _OPTIONGREEKS._serialized_start=360
  _OPTIONGREEKS._serialized_end=482
  _EXTENDEDFEEDDETAILS._serialized_start=485
  _EXTENDEDFEEDDETAILS._serialized_end=804
  _OHLC._serialized_start=806
  _OHLC._serialized_end=914
  _MARKETFULLFEED._serialized_start=917
  _MARKETFULLFEED._serialized_end=1293
  _INDEXFULLFEED._serialized_start=1296
  _INDEXFULLFEED._serialized_end=1484
  _FULLFEED._serialized_start=1487
  _FULLFEED._serialized_end=1662
  _OPTIONCHAIN._serialized_start=1665
  _OPTIONCHAIN._serialized_end=1961
  _FEED._serialized_start=1964
  _FEED._serialized_end=2173
  _FEEDRESPONSE._serialized_start=2176
  _FEEDRESPONSE._serialized_end=2419
  _FEEDRESPONSE_FEEDSENTRY._serialized_start=2330
  _FEEDRESPONSE_FEEDSENTRY._serialized_end=2419
# @@protoc_insertion_point(module_scope)


import requests
import datetime
import json



class ShoonyaHist:
    LTPData_per_sec = 10
    MaxQuoteSymbolLimit = 1
    _timeout = 15

    BASE_URL = "https://api.shoonya.com/NorenWClientTP/"
    ENDPOINTS = { # For V3
        "getData" : "TPSeries",
        "getDataDay" : "EODChartData",
        "getQuotes" : "GetQuotes"
    }

    HistoricalMapping = {
        "1": [1],   # shoonya has 1 minute data ,but no limit on days
        "3": [3],
        "5": [5],
        "10": [10],
        "15": [15],
        "30": [30],
        "60": [60],
        "120": [120],
        "240": [240],
        "1D": ['1D']
    }

    def __init__(self,sessionData : dict):
        self.__susertoken = sessionData['Sessionid']
        self.__username = sessionData['Clientid']
        # self.__accountid = sessionData['username']
        self.reqsession = requests.Session()
        self.__supportedFormats = list(self.info()['supported timeformats']["timeValues"].values())

    def requestAPI(self, method, url, body=None, data=None, timeout=_timeout, output=None, headers=None):
        try:
            if headers:
                resp = self.reqsession.request(method, url, json=body, data=data, timeout=timeout)
            else:
                resp = self.reqsession.request(method, url, json=body, data=data, timeout=timeout, headers=headers)
            data = resp.json()
            if output == 'list' and type(data) == list:
                return {"status": True, "data": data, "error": False, "message": "Data Received"}

            elif data.get("stat") == "Ok":
                return {"status": True, "data": [data], "error": False, "message": "Data Received"}
            else:
                try:
                    return {"status": False, "data": [], "error": True, "message": data['emsg']}
                except :
                    return {"status": False, "data": [], "error": True, "message": data}

        except requests.exceptions.Timeout:
            return {"status": False, "data": [], "error": True, "message": "Timeout error"}
        except Exception as e:
            try:
                return {"status": False, "data": [], "error": True, "message": f"status_code : {resp.status_code}, text : {resp.text}"}
            except:
                return {"status": False, "data": [], "error": True, "message": str(e)}

    def get_quotes(self, symbols: list):
        main_data = {}
        tokenMap = {}
        nonTokens = []

        # Map tokens to exchange:symbol and categorize tokens
        for token in symbols:
            histToken = token.get('histToken')
            if histToken:
                tokenMap[histToken] = token['token']
            else:
                nonTokens.append(token['token'])

        if not tokenMap:
            return {
                "status": False,
                "data": [],
                "error": True,
                "message": "No histToken found for any symbols"
            }

        tokenMapKeys = list(tokenMap.keys())

        for i in range(0, len(tokenMapKeys), self.MaxQuoteSymbolLimit):
            url = f'{self.BASE_URL}{self.ENDPOINTS["getQuotes"]}'
            values              = {}
            values["uid"]       = self.__username
            values["exch"],values["token"] = tokenMapKeys[i].split(", ")
            payload = 'jData=' + json.dumps(values) + f'&jKey={self.__susertoken}'

            try:
                response = self.requestAPI('POST', url, data = payload)
                if not response.get('status') or response.get('error'):
                    return response

                api_data = response.get('data', [])
                if not api_data or not isinstance(api_data[0], dict):
                    return {
                    "status": False,
                    "data": [],
                    "error": True,
                    "message": "No data received"
                    }
                api_data = api_data[0]
                main_data[tokenMap[tokenMapKeys[i]]] = {
                    "timestamp_str": datetime.datetime.fromtimestamp(
                        int(api_data.get('lut', datetime.datetime.now().timestamp()))
                    ).strftime('%Y-%m-%d %H:%M:%S'),
                    "timestamp": str(api_data.get('lut', datetime.datetime.now().timestamp())),
                    "symbol": tokenMap[tokenMapKeys[i]],
                    "ltp": float(api_data.get('lp', 0)),
                    "prev_day_close": float(api_data.get('c', 0)),
                    "oi": 0,
                    "prev_day_oi": 0,
                    "turnover": int(api_data.get('v', 0)),
                    "best_bid_price": float(api_data.get('bp1', 0)),
                    "best_ask_price": float(api_data.get('sp1', 0)),
                    "best_bid_qty": int(api_data.get('bq1', 0)),
                    "best_ask_qty": int(api_data.get('sq1', 0)),
                    "ttq": int(api_data.get('v', 0)),
                    "token": tokenMap[tokenMapKeys[i]]
                }
            except Exception as e:
                return {
                    "status": False,
                    "data": [],
                    "error": True,
                    "message": f"Error occurred: {str(e)}"
                }

        # Fill in empty data for nonTokens
        for token in nonTokens:
            main_data[token] = {}

        return {
            "status": True,
            "data": main_data,
            "error": False,
            "message": "Data Received"
        }

    def to_midnight_timestamp(self, dt):
        return int(dt.replace(hour=0, minute=0, second=0, microsecond=0).timestamp())


    def getHistorical(self, token, startTime : datetime, endTime : datetime, interval : str): #,sec = False
        if interval not in self.__supportedFormats:
            return {"status" : False, "error" : True, "message" : f"Invalid interval. Please input as per the format: {self.info()}", "data" : []}

        if not token:
            return {"status" : False, "data" : [], "error" : True, "message" : "No histToken found for the symbol"}


        startTime = self.to_midnight_timestamp(startTime)
        endTime = min(self.to_midnight_timestamp(endTime), self.to_midnight_timestamp(datetime.datetime.now()))


        if interval == "1D":
            url = f'{self.BASE_URL}{self.ENDPOINTS["getDataDay"]}'
            values              = {}
            values["uid"]       = self.__username
            values["sym"]      = ":".join(token.split(", "))
            values["from"]     = str(startTime)
            values["to"]       = str(endTime)
            headers = {"Content-Type": "application/json; charset=utf-8"}
        else:
            url = f'{self.BASE_URL}{self.ENDPOINTS["getData"]}'
            values              = {'ordersource':'API'}
            values["uid"]       = self.__username
            values["exch"]      = token.split(", ")[0]
            values["token"]     = token.split(", ")[1]
            values["st"] = str(startTime)
            values["en"] = str(endTime)
            values["intrv"] = str(self.HistoricalMapping[interval][0])
            headers = {}

        payload = 'jData=' + json.dumps(values) + f'&jKey={self.__susertoken}'
        response = self.requestAPI('POST', url, data = payload, headers=headers, output='list')

        if not response.get('status') or response.get('error'):
            return response

        candle_data = response.get('data', [])

        if candle_data:
            main_data = self.ConvertToSTD(candle_data)
            main_data = main_data[::-1]
        else:
            return {"status": False, "data": [], "error": True, "message": "Data not available"}

        return {"status": True, "data": main_data, "error": False, "message": "Data Received"} if main_data else \
            {"status": False, "data": [], "error": True, "message": "Data not available"}


    def ConvertToSTD(self, data):
        if not data:
            return []

        std_format = [
            {
                'datetime': datetime.datetime.strptime(candle.get('time'), "%d-%m-%Y %H:%M:%S").strftime("%Y-%m-%d %H:%M:%S"),
                'open': candle.get('into'),
                'high': candle.get('inth'),
                'low': candle.get('intl'),
                'close': candle.get('intc'),
                'volume': int(candle.get('intv', 0)),
                'oi': int(candle.get('oi', 0))
            }
            for candle in data
        ]

        return std_format

    def info(self):
        return {"supported timeformats" :
            {"timeValues" :
                {
                    "1 minute" : "1",
                    "3 minute" : "3",
                    "5 minute"  : "5",
                    "10 minute" : "10",
                    "15 minute" : "15",
                    "30 minute" : "30",
                    "60 minute" : "60",
                    "120 minute" : "120",
                    "240 minute" : "240",
                    "1 Day" : "1D"
                }}}




import datetime
import threading
import json
import traceback
import queue
import time
import os 
from .dependencies import XTSMDSocket_io, XTS_MarketData, SparkLib

class IIFLXTSWS():
    tmdiff = (datetime.datetime(1980, 1, 1, 0, 0, 0) - datetime.datetime(1970, 1, 1, 0, 0, 0)).total_seconds()
    apiURL = 'https://ttblaze.iifl.com/apimarketdata' 
    URL = "https://ttblaze.iifl.com"
    TOKENS = {}
    BROKER = "IIFLXTS"

    def __init__(self, accountData, accessToken = None, dataType = "ltp", 
                 onLtp = None, 
                 onDepth = None, 
                 onError = None, 
                 onClose = None,
                 onOpen = None,
                 ) :
        self.accountData = accountData
        self.dType = 1501 if dataType == "ltp" else 1502 if dataType == "depth" else None #Currently Only supports LTP Updates.
        self.onLtp = self.__onLTP if not onLtp  else onLtp
        self.onDepth = self.__onDepth if not onDepth else onDepth
        self.onError = self.__onError if not onError else onError
        self.onClose = self.__onClose if not onClose else onClose
        self.onOpen = self.__onOpen if not onOpen else onOpen
        self.actionQueue = queue.Queue()
        self.run = True
        if not accessToken :
            if os.environ.get("MATICALGOS_AccessToken"):
                accessToken = os.environ["MATICALGOS_AccessToken"] 
            else: 
                raise Exception("Please generate access token.")
        self.spk = SparkLib(access_token=accessToken)
        
    def getTokens(self, tokens):
        data = self.spk.getBrokerTokens(tokens=tokens, broker=self.BROKER)
        if data['status'] and not data['error'] : 
            tokens = data['data']
            return tokens
        else: 
            raise Exception (data['message'])
    
    def getSymbol(self, exchange, token):
        return self.TOKENS[", ".join([str(exchange), str(token)])]
    
    def onLTPHandler(self, message):
        try: 
            data = json.loads(message)
            touchlineData = data if not data.get("Touchline") else data['Touchline']
            timestamp = self.tmdiff + data['ExchangeTimeStamp'] if data.get("ExchangeTimeStamp") else self.tmdiff + data['LastUpdateTime']  #touchlineData['LastUpdateTime']
            dt = datetime.datetime.utcfromtimestamp(timestamp)
            token = self.getSymbol(data['ExchangeSegment'], data['ExchangeInstrumentID'])
            msg = {"timestamp_str" : str(dt), 
                   "timestamp" : str(int(float(timestamp))),
                   "symbol" : token, 
                   "ltp" : float(touchlineData['LastTradedPrice']),
                   "prev_day_close" : float(touchlineData['Close']), 
                   "oi" : 0,
                   "prev_day_oi" : 0,
                   "turnover" : int(touchlineData['TotalTradedQuantity']), 
                   "best_bid_price" : float(touchlineData['BidInfo']['Price']), 
                   "best_ask_price" : float(touchlineData['AskInfo']['Price']), 
                   "best_bid_qty" : int(touchlineData['BidInfo']['Size']),
                   "best_ask_qty" : int(touchlineData['AskInfo']['Size']), 
                   "ttq" : int(touchlineData['TotalTradedQuantity']),  
                   "token" : token}
            self.onLtp(msg)
                
        except Exception as e : 
            self.onError(f"Error : {e}, Traceback : {traceback.format_exc()}, data : {message}")

    def onDepthHandler(self, message):
        self.onDepth(message)

    def __onLTP(self, message):
        print(message)
    
    def __onDepth(self, message):
        print(message)
    
    def __onError(self, message):
        print("error : ",message)

    def __onClose(self):
        print("Connection Closed")

    def __onOpen(self):
        print("Connection Opened")

    def Subscribe(self, tokens):
        try: 
            notFound = []
            tokens = self.getTokens(tokens)
            Instruments = []
            for tk in tokens:
                wsToken = tk['wsToken']
                if wsToken != None : 
                    exch, tok = wsToken.split(", ")
                    Instruments.append({"exchangeSegment" : exch, 'exchangeInstrumentID' : int(tok)})
                    self.TOKENS[wsToken] = tk['token']
                else: 
                    notFound.append(tk['token'])
            
            if Instruments != [] : 
                response = self.xtconn.subscribe(Instruments, self.dType)
                if response.get('error') :
                    self.onError(f"Error in token Subscription : {response}")
            if notFound != [] : 
                self.onError(f"ERROR : Unable to find tokens for {notFound} ")
        except Exception as e : 
            self.onError(f"Error in token Subscription : {e}, Traceback : {traceback.format_exc()}")

    def Unsubscribe(self, tokens):
        try: 
            notFound = []
            tokens = self.getTokens(tokens)
            Instruments = []
            for tk in tokens:
                wsToken = tk['wsToken']
                if wsToken != None : 
                    exch, tok = wsToken.split(", ")
                    Instruments.append({"exchangeSegment" : exch, 'exchangeInstrumentID' : int(tok)})
                else: 
                    notFound.append(tk['token'])
            
            if Instruments != [] : 
                response = self.xtconn.unsubscribe(Instruments, self.dType)
                if response.get('error') :
                    self.onError(f"Error in token Subscription : {response}")
            if notFound != [] : 
                self.onError(f"ERROR : Unable to find tokens for {notFound} ")
        except Exception as e : 
            self.onError(f"Error in token Unsubscription : {e}, Traceback : {traceback.format_exc()}")

    def on_message(self, data):
        pass
    
    def on_message1505_json_full(self, data):
        pass

    def on_message1507_json_full(self, data):
        pass

    def on_message1510_json_full(self, data):
        pass

    def on_message1512_json_full(self, data):
        pass

    def on_message1105_json_full(self, data):
        pass

    def on_message1501_json_partial(self, data):
        pass

    def on_message1502_json_partial(self, data):
        pass

    def on_message1505_json_partial(self, data):
        pass

    def on_message1510_json_partial(self, data):
        pass

    def on_message1512_json_partial(self, data):
        pass

    def on_message1105_json_partial(self, data):
        pass

    def closeConnection(self):
        self.run = False
        time.sleep(2)
        self.soc.sid.disconnect()
        self.onClose()

    def __connect(self):
        self.soc = XTSMDSocket_io(token=self.accountData['Sessionid'], userID=self.accountData['Clientid'],
                                  reconnection_attempts=5, URL=self.URL)
        self.soc.on_connect = self.onOpen
        self.soc.on_message = self.on_message
        self.soc.on_message1502_json_full = self.onDepthHandler
        self.soc.on_message1505_json_full = self.on_message1505_json_full
        self.soc.on_message1507_json_full = self.on_message1507_json_full
        self.soc.on_message1510_json_full = self.on_message1510_json_full
        self.soc.on_message1501_json_full = self.onLTPHandler
        self.soc.on_message1512_json_full = self.on_message1512_json_full
        self.soc.on_message1105_json_full = self.on_message1105_json_full
        self.soc.on_message1502_json_partial = self.on_message1502_json_partial
        self.soc.on_message1505_json_partial = self.on_message1505_json_partial
        self.soc.on_message1510_json_partial = self.on_message1510_json_partial
        self.soc.on_message1501_json_partial = self.on_message1501_json_partial
        self.soc.on_message1512_json_partial = self.on_message1512_json_partial
        self.soc.on_message1105_json_partial = self.on_message1105_json_partial
        self.soc.on_disconnect = self.onClose
        self.soc.on_error = self.onError
        el = self.soc.get_emitter()
        el.on('connect', self.onOpen)
        el.on('1501-json-full', self.onLTPHandler)
        el.on('1502-json-full', self.onDepthHandler)
        el.on('1507-json-full', self.on_message1507_json_full)
        el.on('1512-json-full', self.on_message1512_json_full)
        el.on('1105-json-full', self.on_message1105_json_full)
        try: 
            self.soc.connect()
        except Exception as e :  
            self.onClose(f"Error : {e}, Traceback : {traceback.format_exc()}")
    
    def connect(self):
        self.xtconn = XTS_MarketData(sessionid = self.accountData['Sessionid'], URL=self.apiURL)
        threading.Thread(target=self.__connect).start()




from typing import Optional, Callable
import asyncio
import json
import ssl
import requests
import websockets
from google.protobuf.json_format import MessageToDict
from . import MarketDataFeedUps as pb

def decode_protobuf(buffer):
    """Decode protobuf message."""
    feed_response = pb.FeedResponse()
    feed_response.ParseFromString(buffer)
    return feed_response


class AsyncUpstockOrderUpdate:
    Authorize_URL = 'https://api.upstox.com/v2/feed/market-data-feed/authorize'
    Api_Version = '2.0'

    def __init__(
            self,
            auth_token: str,
            on_message: Optional[Callable] = None,
            on_error: Optional[Callable] = None,
            on_connect: Optional[Callable] = None,
            on_close: Optional[Callable] = None,
            reconnect_retry: int = 5,
    ):
        self.auth_token = f'Bearer {auth_token}'
        self.onmessage = on_message
        self.onerror = on_error
        self.onconnect = on_connect
        self.onclose = on_close
        self.max_retry_attempts = reconnect_retry
        self.current_retry_attempt = 0
        self.websocket = None
        self.running = True  # Flag to control running state

    def getWSUrl(self):
        headers = {
            "Authorization": self.auth_token,
            "Api-Version": self.Api_Version,
            "Accept": "application/json"
        }
        response = requests.get(self.Authorize_URL, headers=headers)
        if response.status_code == 200:
            return response.json().get('data').get('authorized_redirect_uri')
        else:
            return None


    async def connect(self):
        """Connect to the WebSocket and handle messages asynchronously."""
        try:

            # Extract WebSocket URL
            ws_url = self.getWSUrl()
            # Set up SSL context
            ssl_context = ssl.create_default_context()
            ssl_context.check_hostname = False
            ssl_context.verify_mode = ssl.CERT_NONE

            async with websockets.connect(ws_url, ssl=ssl_context) as websocket:
                self.websocket = websocket
                if self.onconnect:
                    await self.onconnect()
                self.current_retry_attempt = 0
                # crate a new task to listen for messages
                asyncio.create_task(self.listen())

                await asyncio.Future()
        except Exception as e:
            if self.onerror:
                await self.onerror(e)
            await self.retry_connect()

    async def retry_connect(self):
        """Attempt to reconnect with exponential backoff."""
        if self.current_retry_attempt < self.max_retry_attempts and self.running:
            retry_delay = 2 ** self.current_retry_attempt  # Exponential backoff
            print(f"Retrying connection in {retry_delay} seconds ", f"({self.current_retry_attempt + 1}/{self.max_retry_attempts})")
            await asyncio.sleep(retry_delay)
            self.current_retry_attempt += 1
            await self.connect()
        else:
            self.running = False
            await self.onclose(1006, "Connection closed due to max retry attempts")

    async def listen(self):
        """Listen for messages from the WebSocket."""
        try:
            while self.running:
                message = await self.websocket.recv()
                decoded_data = decode_protobuf(message)
                data_dict = MessageToDict(decoded_data)

                if self.onmessage:
                    await self.onmessage(data_dict)

        except websockets.ConnectionClosed as e:
            if self.onclose:
                await self.onclose(e.code, e.reason)
            await self.retry_connect()
        except Exception as e:
            if self.onerror:
                await self.onerror(e)

    def subscribe(self, mode: str, instrument_keys: list):
        """Send a subscription message to the WebSocket."""
        try:
            data = {
                "guid": "someguid",
                "method": "sub",
                "data": {
                    "mode": mode,
                    "instrumentKeys": instrument_keys
                }
            }
            binary_data = json.dumps(data).encode('utf-8')
            asyncio.create_task(self.websocket.send(binary_data))
            # self.websocket.send(binary_data)
        except Exception as e:
            print(f"Error in subscribing: {e}")


    def unsubscribe(self, mode: str, instrument_keys: list):
        """Send an unsubscription message to the WebSocket."""
        try:
            data = {
                "guid": "someguid",
                "method": "unsub",
                "data": {
                    "mode": mode,
                    "instrumentKeys": instrument_keys
                }
            }
            binary_data = json.dumps(data).encode('utf-8')
            # self.websocket.send(binary_data)
            asyncio.create_task(self.websocket.send(binary_data))
        except Exception as e:
            print(f"Error in unsubscribing: {e}")

    async def close_connection(self):
        """Close the WebSocket connection."""
        self.running = False  # Stop listening loop
        if self.websocket:
            try:
                await self.websocket.close()
            except Exception as e:
                print(f"Error closing connection: {e}")
        else:
            print("Connection already closed.")

